## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""Solvers for the Schrödinger/Lindblad equation

The code may use the qutip solvers, or use some internal solvers
that may be faster in some circumstances.


"""

import sys
import numpy as np
from qutip import *
from qocttools.math_extra import rk4
from scipy.linalg import expm


def solve_h_is_function(solve_method, H, f, y0, time,
                        returnQoutput = True,
                        options = None,
                        interaction_picture = False):
    """Solver for the case in which the hamiltonian is given as a function
    """
    H0 = H.H0
    V = H.V
    cops = H.A

    # What are we propagating?
    obj = None
    if y0.type == 'oper':
        if cops is not None:
            obj = 'density'
        else:
            obj = 'propagator'
    else:
        obj = 'state'
    if obj is None:
        raise Exception('Do not know what is the object to propagate')

    if not isinstance(V, list):
        V = [V]
    if not isinstance(f, list):
        f = [f]

    if obj == 'propagator':
        if not y0.isunitary:
            q_, r_ = np.linalg.qr(y0.full())
            y0 = Qobj(q_)
            r = Qobj(r_)

    args = { "f": [f[l].fu for l in range(len(f))] }
    result = sesolve(H.H0, y0, time, options = options, args = args)
    if returnQoutput:
        if obj == 'propagator' and 'r' in locals():
            raise Exception("Internal Error")
        return result
    else:
        dim = H.H0(0.0, args).shape[0]
        if y0.type == 'oper':
            output = np.zeros([time.size, dim, dim], dtype = complex)
            for j in range(time.size):
                if obj == 'propagator' and 'r' in locals():
                    output[j, :, :] = result.states[j].full() @ r_
                else:
                    output[j, :, :] = result.states[j].full()
        else:
            output = np.zeros([time.size, dim], dtype = complex)
            for j in range(time.size):
                output[j, :] = result.states[j].full()[:, 0]
        return output


def solve(solve_method, H, f, y0, time,
          returnQoutput = True,
          options = None,
          interaction_picture = False):
    """
    solve(solve_method, H, f, u, y0, time)
    
    This function propagate the object y0 using either the Schrödinger equation 
    with H = H0 + f(t)*V as Hamiltonian, or Lindblad's equation. If an inverse time array is given, the solver
    take y0 as the state at final time, making a backward propagation.

    The problem may be solved in the interaction picture if interaction_picture = True.
    Note that in that case H0 should be diagonal on entry (the system should be represented
    in the eigenbasis of H0). In this case, the output will also be in the interaction 
    picture.
    
    Parameters
    ----------
    solve_method: string indicating the propagation method
    H0: hamiltonian
        Hamiltonian's time independet part 
    V:  list of functions
        Hamiltonian's perturbation component 
    f:  list of pulse
       pulse class defined in typical_pulse.py based on Fourier expansion 
       parametrization
    y0: 
       initial state
    time: ndarray
       array that contain each time step.
    u: ndarray
       array that contain the control parameters of the pulse, needed to
       build the args diccionary defined in pulse class
           
    Returns
    .......
    qutip.result:
        qutiip.result type object with the propagation data 
    """

    if H.function:
        return solve_h_is_function(solve_method, H, f, y0, time,
                                   returnQoutput,
                                   options,
                                   interaction_picture)

    H0 = H.H0
    V = H.V
    cops = H.A


    # What are we propagating?
    obj = None
    if y0.type == 'oper':
        if cops is not None:
            obj = 'density'
        else:
            obj = 'propagator'
    else:
        obj = 'state'
    if obj is None:
        raise Exception('Do not know what is the object to propagate')

    if not isinstance(V, list):
        V = [V]
    if not isinstance(f, list):
        f = [f]

    if solve_method == 'sesolve':

        if obj == 'propagator':
            if not y0.isunitary:
                q_, r_ = np.linalg.qr(y0.full())
                y0 = Qobj(q_)
                r = Qobj(r_)

        if interaction_picture:
            vsch = []
            for vop in V:
                vsch.append(vop.full())
            def Hamiltonianfunc(t, Hargs):
                V = Hargs["V"]
                e = Hargs["eigenvalues"]
                dim = e.shape[0]
                Ht = Qobj(np.zeros([dim, dim]))
                ft = [f[l].fu(t) for l in range(len(f))]
                gt = [H.g(ft, l) for l in range(len(V))]
                for j in range(len(V)):
                    v = vsch[j].copy()
                    for m in range(dim):
                        for n in range(dim):
                            v[m, n] = np.exp(-1j*t*(e[n]-e[m])) * v[m, n]
                    Ht = Ht + gt[j] * Qobj(v)
                return Ht
            Hargs = {'V': V, 'eigenvalues': H0.diag()}
            if cops is None:
                result = sesolve(Hamiltonianfunc, y0, time, args = Hargs)
            else:
                result = mesolve(Hamiltonianfunc, y0, time, cops, args = Hargs, options = options)
        else:
            Hamiltonian = [H0]
            for j in range(len(V)):
                def make_f(j):
                    def func(t, args):
                        ft = [f[l].fu(t) for l in range(len(f))]
                        return H.g(ft, j)
                    return func
                Hamiltonian.append([V[j], make_f(j)])
            if options == None:
                options = Options()
            if cops is None:
                result = sesolve(Hamiltonian, y0, time, options = options)
            else:
                result = mesolve(Hamiltonian, y0, time, cops, options = options)
        if returnQoutput:
            if obj == 'propagator' and 'r' in locals():
                raise Exception("Internal Error")
            return result
        else:
            dim = H0.shape[0]
            if y0.type == 'oper':
                output = np.zeros([time.size, dim, dim], dtype = complex)
                for j in range(time.size):
                    if obj == 'propagator' and 'r' in locals():
                        output[j, :, :] = result.states[j].full() @ r_
                    else:
                        output[j, :, :] = result.states[j].full()
            else:
                output = np.zeros([time.size, dim], dtype = complex)
                for j in range(time.size):
                    output[j, :] = result.states[j].full()[:, 0]
            return output
        
    elif solve_method == 'rk4':
        return rk4solver(H, f, y0, time,
                         returnQoutput = returnQoutput, 
                         interaction_picture = interaction_picture)

    elif solve_method == 'cfmagnus2':
        return cfmagnus2solver(H, f, y0, time,
                               returnQoutput = returnQoutput,
                               interaction_picture = interaction_picture,
                               cops = cops)

    elif solve_method == 'cfmagnus4':
        return cfmagnus4solver(H, f, y0, time,
                               returnQoutput = returnQoutput,
                               interaction_picture = interaction_picture,
                               cops = cops)


def solvep(solve_method, H, f, y0, time,
           returnQoutput = None,
           options = None,
           interaction_picture = None,
           nprocs = 1,
           comm = None):

    if comm == None:
        nprocs = 1
        rank = 0
    else:
        nprocs = comm.Get_size()
        rank = comm.Get_rank()
    nelements = len(y0)

    st_ = np.empty(nprocs, dtype = 'int')
    for j in range(nprocs):
        st_[j] = int(np.floor(nelements * j / nprocs))
    st = int(np.floor(nelements * rank / nprocs))

    length_ = np.empty(nprocs, dtype = 'int')
    for j in range(nprocs):
        length_[j] = int(np.floor(nelements * (j+1) / nprocs)) - st_[j]
    length = int(np.floor(nelements * (rank+1) / nprocs)) - st

    end_ = np.empty(nprocs, dtype = 'int')
    for j in range(nprocs):
        end_[j] = st_[j] + length_[j]-1
    end = st + length - 1


    if nprocs > 1:
        result = [None] * nelements
        for j in range(st, end + 1):
            result[j] = solve(solve_method, H[j], f, y0[j], time,
                              returnQoutput = returnQoutput,
                              options = options,
                              interaction_picture = interaction_picture)
        comm.Barrier()
        for k in range(nprocs):
            for j in range(st_[k], end_[k]+1):
                result[j] = comm.bcast(result[j], root = k)
        return result

    else:
        result = []
        for j in range(len(y0)):
            result.append(solve(solve_method, H[j], f, y0[j], time,
                                returnQoutput = returnQoutput,
                                options = options,
                                interaction_picture = interaction_picture))
        return result


def op(M, cops, y, costate_prop = False):
    if cops is None:
        return np.matmul( -1j * M, y)
    else:
        opy = - 1j * np.matmul(M, y) + 1j * np.matmul(y, M)
        if not costate_prop:
            for j in range(len(cops)):
                cops_ = cops[j].full()
                copsd_ = (cops[j].dag()).full()
                opy = opy + np.matmul(np.matmul(cops_, y), copsd_) \
                      - 0.5 * np.matmul(np.matmul(y, copsd_), cops_) \
                      - 0.5 * np.matmul(np.matmul(copsd_, cops_), y)
        else:
            for j in range(len(cops)):
                cops_ = cops[j].full()
                copsd_ = (cops[j].dag()).full()
                opy = opy - np.matmul(np.matmul(copsd_, y), cops_) \
                      + 0.5 * np.matmul(np.matmul(y, copsd_), cops_) \
                      + 0.5 * np.matmul(np.matmul(copsd_, cops_), y)
        return opy


def exppsi(M, cops, dt, y, order = 4):
    opy = y.copy()
    expy = y.copy()
    factor = 1.0
    for i in range(1, order+1):
        factor = factor/i
        if dt < 0.0:
            opy = dt * op(M, cops, opy, costate_prop = True)
        else:
            opy = dt * op(M, cops, opy, costate_prop = False)
        expy = expy + factor * opy
    return expy


def intoper(v, e, t):
    dim = v.shape[0]
    m = np.kron( np.exp(1j * t * e).reshape(dim, 1), np.exp(-1j * t * e).reshape(1, dim) )
    vi = v * m
    return vi


def cfmagnus2solver(H, f, y0, time,
                    returnQoutput = True,
                    interaction_picture = False,
                    cops = None):
    """ Implementation of the exponential midpoint rule.

    It is also the second-order commutator-free Magnus method.
    """
    H0 = H.H0
    V = H.V

    if type(H0) is not qutip.qobj.Qobj:
        raise TypeError

    h0 = H0.full()
    dt = time[1]-time[0]
    dim = H0.shape[0]
    if returnQoutput:
        output = solver.Result()
        output.solver = 'cfmagnus2'
        output.times = time
        output.states = []
        output.states.append(y0)
        if y0.type == 'ket' or y0.type == 'bra':
            y = y0.full()[:, 0]
        else:
            y = y0.full()
    else:
        if y0.type == 'oper':
            output = np.zeros([time.size, dim, dim], dtype = complex)
            output[0, :, :] = y0.full()
            y = y0.full()
        else:
            output = np.zeros([time.size, dim], dtype = complex)
            output[0, :] = y0.full()[:, 0]
            y = y0.full()[:, 0]

    for j in range(time.size-1):
        t = time[j] + 0.5*dt
        ft = [ f[k].fu(t) for k in range(len(f)) ]
        g = [ H.g(ft, k) for k in range(len(V)) ]
        if interaction_picture:
            M = np.zeros_like(h0)
            for k in range(len(V)):
                vi = intoper(V[k].full(), np.diag(h0), t)
                M = M + g[k] * vi
        else:
            M = h0.copy()
            for k in range(len(V)):
                M = M + g[k] * V[k].full()
        y = exppsi(M, cops, dt, y)
        if returnQoutput:
            youtput = y.copy()
            output.states.append(Qobj(youtput))
        else:
            output[j+1, :] = y[:]

    return output


def cfmagnus4solver(H, f, psi0, time,
                    returnQoutput = True,
                    interaction_picture = False, 
                    cops = None):
    """ Implementation of the fourth order commutator-free Magnus method.
    """
    H0 = H.H0
    V = H.V
    v = [V[j].full() for j in range(len(V))]

    if type(H0) is not qutip.qobj.Qobj:
        raise TypeError

    h0 = H0.full()
    dt = time[1]-time[0]
    dim = H0.shape[0]

    if returnQoutput:
        output = solver.Result()
        output.solver = 'cfmagnus4'
        output.times = time
        output.states = []
        output.states.append(psi0)
        if psi0.type == 'ket' or psi0.type == 'bra':
            psi = psi0.full()[:, 0]
        else:
            psi = psi0.full()
    else:
        if psi0.type == 'oper':
            output = np.zeros([time.size, dim, dim], dtype = complex)
            output[0, :, :] = psi0.full()
            psi = psi0.full()
        else:
            output = np.zeros([time.size, dim], dtype = complex)
            output[0, :] = psi0.full()[:, 0]
            psi = psi0.full()[:, 0]

    a1 = (3.0-2.0*np.sqrt(3.0))/12.0
    a2 = (3.0+2.0*np.sqrt(3.0))/12.0
    c1 = 0.5 - np.sqrt(3.0)/6.0
    c2 = 0.5 + np.sqrt(3.0)/6.0

    for j in range(time.size-1):
        t1 = time[j] + c1*dt
        t2 = time[j] + c2*dt
        ft1 = [ f[k].fu(t1) for k in range(len(f)) ]
        ft2 = [ f[k].fu(t2) for k in range(len(f)) ]
        g1 = [ H.g(ft1, k) for k in range(len(V)) ]
        g2 = [ H.g(ft2, k) for k in range(len(V)) ]
        if interaction_picture:
            # We will assume that H0 is diagonal on entry.
            M1 = np.zeros_like(h0)
            M2 = np.zeros_like(h0)
            for k in range(len(V)):
                vi1 = intoper(v[k], np.diag(h0), t1)
                vi2 = intoper(v[k], np.diag(h0), t2)
                M1 = M1 + a1 * g1[k] * vi1 + a2 * g2[k] * vi2
                M2 = M2 + a2 * g1[k] * vi1 + a1 * g2[k] * vi2

        else:
            M1 = (a1 + a2) * h0
            M2 = (a1 + a2) * h0
            for k in range(len(V)):
                M1 = M1 + (a1 * g1[k] + a2 * g2[k]) * v[k]
                M2 = M2 + (a2 * g1[k] + a1 * g2[k]) * v[k]

        M = M2
        psi = exppsi(2*M, cops, dt/2, psi)
        M = M1
        psi = exppsi(2*M, cops, dt/2, psi)

        if returnQoutput:
            output.states.append(Qobj(psi))
        else:
            if psi0.type == 'oper':
                output[j+1, :, :] = psi[:, :]
            else:
                output[j+1, :] = psi[:]


    return output


def rk4solver(H, f, psi0, time, returnQoutput = True,
             interaction_picture = False):
    # WARNING: This may not work with operators, specially in the interaction picture.
    H0 = H.H0
    V = H.V
    if type(H0) is not qutip.qobj.Qobj:
        raise TypeError

    h0 = H0.full()
    dt = time[1]-time[0]
    dim = H0.shape[0]
    if returnQoutput:
        output = solver.Result()
        output.solver = 'rk4'
        output.times = time
        output.states = []
        output.states.append(psi0)
    else:
        if psi0.type == 'oper':
            output = np.zeros([time.size, dim, dim], dtype = complex)
            output[0, :, :] = psi0.full()
        else:
            output = np.zeros([time.size, dim], dtype = complex)
            output[0, :] = psi0.full()[:, 0]

    def dynfun(t, xi):
        ft = [ f[k].fu(t) for k in range(len(f)) ]
        g = [ H.g(ft, k) for k in range(len(V)) ]
        if interaction_picture:
            y = np.zeros_like(xi)
            for j in range(len(V)):
                vi = intoper(V[j].full(), np.diag(h0), t)
                y = y - 1j * g[j] * np.matmul(vi, xi)
            return y
        else:
            y = -1j * np.matmul(h0, xi)
            for j in range(len(V)):
                y = y - 1j * g[j] * np.matmul(V[j].full(), xi)
            return y

    if psi0.type == 'oper':
        xi = psi0.full()
    else:
        xi = psi0.full()[:, 0]

    for j in range(time.size-1):
        xi = rk4(xi, dynfun, time[j], dt)
        if returnQoutput:
            output.states.append(Qobj(xi))
        else:
            if psi0.type == 'oper':
                output[j+1, :, :] = xi[:, :]
            else:
                output[j+1, :] = xi[:]
 
    return output



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module holds a hBN model


"""

import numpy as np
import qutip as qt

#a  =  2.5  #A
a = 4.7243153 # a.u.
#c  =  6.6  #A
c = 12.472192 # a.u.
#eb =  2.46 #eV
eb = 0.090403333 # a.u.
#en = -2.55 #eV
en = -0.093710772 # a.u.

#gam =[-2.16,-0.04,-0.08] # eV
gam =[-0.093710772, -0.0014699729, -0.0029399458] # eV

a1 = a * np.array([np.sqrt(3.0)/2.0,  1.0/2.0])
a2 = a * np.array([0.0, 1.0])

delta1 = np.zeros([3, 2])
delta2 = np.zeros([6, 2])
delta3 = np.zeros([3, 2])

d1 = (1.0/3.0) * np.sqrt(3) * a
d2 = a
#d3 = 2*a / ( np.sqrt(2.0)*np.sqrt(1-np.cos(np.pi/3)) )
d3 = d1 * (1+2*np.cos(np.pi/3))

for j in range(3):
    theta = np.pi/3 + j*(2.0*np.pi/3)
    delta1[j, :] = d1 * np.array([np.cos(theta), np.sin(theta)])

for j in range(6):
    theta = np.pi/6 + j*(2.0*np.pi/6)
    delta2[j, :] = d2 * np.array([np.cos(theta), np.sin(theta)])

for j in range(3):
    theta = j*(2.0*np.pi/3)
    delta3[j, :] = d3 * np.array([np.cos(theta), np.sin(theta)])


#print(d1, d2, d3)

kd = 4.0 * np.pi / (np.sqrt(3) * a)

b1 = kd * np.array([1.0, 0.0])
b2 = kd * np.array([-1.0/2.0, np.sqrt(3.0)/2.0])

Gammapoint = np.array([0.0, 0.0])
Kpoint  = kd * np.array([0.5, 1.0/(2.0*np.sqrt(3.0))])
Mpoint = kd * np.array([0.5, 0.0])

#Path in regiprocal space
dir=np.array([2/3,1/3])
kgrid = np.linspace(-3.5, 3.5, 101)[:,None]*dir[None,:]
klen  = np.sign(kgrid[:,0])*np.sqrt((kgrid**2).sum(axis=1))

def ffk(kgrid):
    #fk = np.exp(-1j*a*kgrid[0]/np.sqrt(3.)) + 2.*np.exp(1j*a*kgrid[0]/(2.*np.sqrt(3.)))*np.cos(a/2.*kgrid[1])
    fk = 0.0
    for j in range(3):
        fk = fk + np.exp(1j * (kgrid[0] * delta1[j, 0] + kgrid[1] * delta1[j, 1]) )
    return fk

def fgk(kgrid):
    #gk = 2.*np.cos(a*kgrid[1]) + 2.*np.cos(a*np.sqrt(3.)/2.*kgrid[0]+ a/2.*kgrid[1]) \
    #                             + 2.*np.cos(a*np.sqrt(3.)/2.*kgrid[0]- a/2.*kgrid[1])
    gk = 0.0
    for j in range(6):
        gk = gk + np.exp(1j * (kgrid[0] * delta2[j, 0] + kgrid[1] * delta2[j, 1]) )
    return gk

def fhk(kgrid):
    #hk = np.exp(1j*2.*a*kgrid[0]/np.sqrt(3.)) + 2.*np.exp(-1j*a*kgrid[0]/np.sqrt(3.))*np.cos(a*kgrid[1])
    hk = 0.0
    for j in range(3):
        hk = hk + np.exp(1j * (kgrid[0] * delta3[j, 0] + kgrid[1] * delta3[j, 1]) )
    return hk

def dfk(kgrid):
    dfk = np.zeros(2, dtype = complex)
    for j in range(3):
        dfk = dfk + np.exp(1j * (kgrid[0] * delta1[j, 0] + kgrid[1] * delta1[j, 1]) ) * delta1[j, :]
    return 1j * dfk

def dgk(kgrid):
    dgk = np.zeros(2, dtype = complex)
    for j in range(6):
        dgk = dgk + np.exp(1j * (kgrid[0] * delta2[j, 0] + kgrid[1] * delta2[j, 1]) ) * delta2[j, :]
    return 1j * dgk

def dhk(kgrid):
    dhk = np.zeros(2, dtype = complex)
    for j in range(3):
        dhk = dhk + np.exp(1j * (kgrid[0] * delta3[j, 0] + kgrid[1] * delta3[j, 1]) ) * delta3[j, :]
    return 1j * dhk


def H0k(kvec):
    h0kmatrix = np.zeros([2, 2], dtype = complex)
    h0kmatrix[0, 0] = eb + gam[1] * fgk(kvec)
    h0kmatrix[0, 1] = gam[0] * ffk(kvec)  + gam[2] * fhk(kvec)
    h0kmatrix[1, 0] = np.conj(h0kmatrix[0, 1])
    h0kmatrix[1, 1] = en + gam[1] * fgk(kvec)
    return qt.Qobj(h0kmatrix)

def Hxk(kvec):
    hxkmatrix = np.zeros([2, 2], dtype = complex)
    hxkmatrix[0, 0] = gam[1] * dgk(kvec)[0]
    hxkmatrix[0, 1] = gam[0] * dfk(kvec)[0]  + gam[2] * dhk(kvec)[0]
    hxkmatrix[1, 0] = np.conj(hxkmatrix[0, 1])
    hxkmatrix[1, 1] = gam[1] * dgk(kvec)[0]
    return qt.Qobj(hxkmatrix)

def Hyk(kvec):
    hykmatrix = np.zeros([2, 2], dtype = complex)
    hykmatrix[0, 0] = gam[1] * dgk(kvec)[1]
    hykmatrix[0, 1] = gam[0] * dfk(kvec)[1]  + gam[2] * dhk(kvec)[1]
    hykmatrix[1, 0] = np.conj(hykmatrix[0, 1])
    hykmatrix[1, 1] = gam[1] * dgk(kvec)[1]
    return qt.Qobj(hykmatrix)





## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module contains all the procedures needed to do Floquet optimization.
"""

import qutip as qt
import numpy as np
import scipy as scp
import scipy.linalg as la
import qocttools.pulses as pulses
import qocttools.cythonfuncs as cyt
from qocttools.math_extra import diff_ridders


def epsi(UT, T):
    """
    Given a matrix UT, returns the Floquet quasienergies associated to it.

    In principle, the matrix should be unitary (or otherwise, there is no
    guarantee that the matrix will have unit eigenvalues). However, the routine
    accepts non-unitary matrix, as it computes the Schur decomposition, and it uses
    the diagonal elements of the Schur form. If the matrix is indeed unitary, the 
    output will be proper Floquet quasienergies.

    Parameters
    ----------
    UT : ndarray
        A unitary matrix, representing an evolution operator.
    T : float
        The Floquet period

    Returns
    -------
    ndarray:
        The Floquet quasienergies (ordered, and in the Floquet Brillouin zone)
    """
    # In principle, one needs the eigenvalues. But what if UT cannot be diagonalized?
    # We will use then the Schur decomposition, and use the diagonal elements of the
    # Schur matrix.
    #evals, evecs = la.eig(UT)
    dim = UT.shape[0]
    S, Z = la.schur(UT)
    evals = np.zeros(dim, dtype = complex)
    for l in range(dim):
        evals[l] = S[l, l]
    eargs = np.angle(evals)
    eargs += (eargs <= -np.pi) * (2 * np.pi) + (eargs > np.pi) * (-2 * np.pi)
    epsilon = -eargs / T
    return np.sort(epsilon)


def depsi(UT, T, delta = 1.0e-4):
    """
    Given a unitary UT, returns the derivatives of the Floquet quasienergies
    with respect to it. Specifically, it returns the derivative of each
    Floquet quasienergy with respect to each matrix element :math:`U_{ij}^*`
    (technically, a Wirtinger derivative, with respect to the complex conjugate
    of the parameter).

    Parameters
    ----------
    UT : ndarray
        A unitary matrix, representing an evolution operator.
    T : float
        The Floquet period
    delta : float, default = 1.0e-4
        The small displacement used to compute the derivatives
        through a finite-difference formula.

    Returns
    -------
    ndarray :
        The numpy array containing the derivatives; the shape is
        (dim, dim, dim): depsi[alpha, :, :] is the matrix of derivatives
        of the alpha-th quasienergy with respect to all the elements of
        the unitary.
    """
    dim = UT.shape[0]
    depsi_ = np.empty([dim, dim, dim], dtype = complex)
    for i in range(dim):
        for j in range(dim):
            UTp = UT.copy()
            UTp[i, j] = UTp[i, j] + delta
            epsilonp = epsi(UTp, T)
            UTm = UT.copy()
            UTm[i, j] = UTm[i, j] - delta
            epsilonm = epsi(UTm, T)
            dx = (epsilonp-epsilonm) / (2*delta)

            UTp = UT.copy()
            UTp[i, j] = UTp[i, j] + 1j*delta
            epsilonp = epsi(UTp, T)
            UTm = UT.copy()
            UTm[i, j] = UTm[i, j] - 1j*delta
            epsilonm = epsi(UTm, T)
            dy = (epsilonp-epsilonm) / (2*delta)

            depsi_[:, i, j] = 0.5 * (dx + 1j*dy)
    return depsi_


def epsilon(H, T, args = None):
    """Computes Floquet quasienergies corresponding to the Hamiltonian H and the period T.

    For this to make sense, the Hamiltonian time-dependence must be periodic. The Hamiltonian
    must be given as a "qutip Hamiotonian", i.e. anything that could be accepted by the qutip
    propagation functions as valid Hamiltonians.

    Parameters
    ----------
    H : qutip Hamiltonian
        The (periodic) Hamiltonian
    T : float
        The period
    args : dict, default = None
        An "args" argument that can b passed to the qutip sesolve function that is used
        to compute the unitary evolution operator at time T.

    Returns
    -------
    ndarray:
        The ordered set of quasienergies.

    """
    options = qt.Options(nsteps = 10000)
    #U = qt.propagxator(H, T, args = args, options = options)
    ntsteps = 5
    dim = 3 # WARNING: This is hard-coded here, but it should not be.
    U = (qt.sesolve(H, qt.qeye(dim), np.linspace(0, T, ntsteps), options = options, args = args)).states[-1]
    ualpha, epsilon = qt.floquet_modes(H, T, U = U)
    idx = epsilon.argsort()
    return epsilon[idx]


def epsilon3(H, f, u, T):
    """
    Given an Hamiltonian H (:class:`hamiltonians.hamiltonian`), a set of driving periodic
    perturbations f, returns the Floquet quasienergies

    Parameters
    ----------
    H : :class:`hamiltonians.hamiltonian`
        The Hamiltonian of the system, given as a hamiltonian class object.
    f : list of :class:`pulses.pulse`
        A list of periodic drivings -- the number should match the number of
       perturbations of the Hamiltonian
    u : ndarray
        The control parameters
    T : float
        The period

    Returns
    -------
    ndarray
        An array with the Floquet quasienergies
    """
    pulses.pulse_collection_set_parameters(f, u)
    if H.function:
        args = { "f": [f[l].fu for l in range(len(f))] }
        H_ = H.H0
    else:
        args = None
        H_ = [H.H0]
        k = 0
        for V in H.V:
            H_.append([V, f[k].f])
            k = k + 1
    return epsilon(H_, T, args = args)


def gradepsilon(H, f, u, T):
    """
    Returns the gradient of the Floquet quasienergies with respect to the control parameters
    of the set of periodic drivings.

    Parameters
    ----------
    H : :class:`hamiltonians.hamiltonian`
        The Hamiltonian of the system, given as a hamiltonian class object.
    f : list of :class:`pulses.pulse`
        A list of periodic drivings -- the number should match the number of
       perturbations of the Hamiltonian
    u : ndarray
        The control parameters
    T : float
        The period

    Returns
    -------
    ndarray
        An array shaped (dim, nu) where dim is the Hilert space dimension, and
        nu is the number of control parameters, containing the derivatives of the
        Floquet quasienergies with respect to those parameters.
    """
    dim = H.dim
    pulses.pulse_collection_set_parameters(f, u)

    if H.function:
        options = qt.Options(nsteps = 10000)
        args = { "f": [f[l].fu for l in range(len(f))] }
        H_ = H.H0
        ntsteps = 5
        U = (qt.sesolve(H_, qt.qeye(dim), np.linspace(0, T, ntsteps), options = options, args = args)).states[-1]
    else:
        args = None
        H0 = H.H0
        Vs = H.V
        H_ = [H.H0]
        k = 0
        for V in H.V:
            H_.append([V, f[k].f])
            k = k + 1
        options = qt.Options(nsteps = 10000)
        U = qt.propagator(H_, T, options = options)

    ualpha, epsilon = qt.floquet_modes(H_, T, U = U)
    idx = epsilon.argsort()
    epsilon = epsilon[idx]
    ualpha_ = [ualpha[j] for j in idx]
    ualpha = ualpha_
    times = np.linspace(0, T, 100)
    dt = times[1]
    ualphat = qt.floquet.floquet_modes_table(ualpha, epsilon, times, H_, T, args = args)

    nu = 0
    for ip in range(len(f)):
        nu = nu + f[ip].nu
    res = np.zeros((dim, nu))

    if H.function:
       Vs_ = []
       for ip in range(len(f)):
           Vs_.append( np.empty(times.shape[0], dtype = qt.Qobj) )
       for j in range(times.shape[0]):
           t = times[j]
           for ip in range(len(f)):
               Vs_[ip][j] = H.V[ip](times[j], args)

    m = 0
    for ip in range(len(f)):
        ft = f[ip]
        if H.function:
            V = Vs_[ip]
        else:
            V = Vs[ip]
        for k in range(f[ip].nu):
            if H.function:
                for alpha in range(dim):
                    res[alpha, m] = 0.5 * (qt.expect(V[0], ualphat[0][alpha]) * ft.dfu(times[0], k) )
                    for j in range(1, times.shape[0]-1):
                        res[alpha, m] = res[alpha, m] + (qt.expect(V[j], ualphat[j][alpha]) * ft.dfu(times[j], k) )
                    res[alpha, m] = res[alpha, m] + 0.5 * (qt.expect(V[j], ualphat[-1][alpha]) * ft.dfu(times[-1], k) )
            else:
                for alpha in range(dim):
                    res[alpha, m] = 0.5 * (qt.expect(V, ualphat[0][alpha]) * ft.dfu(times[0], k) )
                    for j in range(1, times.shape[0]-1):
                        res[alpha, m] = res[alpha, m] + (qt.expect(V, ualphat[j][alpha]) * ft.dfu(times[j], k) )
                    res[alpha, m] = res[alpha, m] + 0.5 * (qt.expect(V, ualphat[-1][alpha]) * ft.dfu(times[-1], k) )
            m = m + 1

    return res * dt/T


def vectortodm(rhovec):
    """
    Transforms a numpy array of dimension (d^2) into an "operator" Qobj object; this object has
    dimension [[d], [d]]. The input 1D array is supposed to store the operator with stacked
    columns.

    Parameters
    ----------
    rhovec : ndarray
        A one-dimensional numpy complex array with d^2 elements

    Returns
    -------
        A [[d], [d]] shaped qutip Qobj
    """
    dim = round(np.sqrt(rhovec.shape[0]))
    return qt.vector_to_operator(qt.Qobj(rhovec, dims = [[[dim], [dim]], [1]]))


def dmtovector(rho):
    """
    Transforms a Qobj [[d], [d]] shaped operator  object into a [[d^2], [1]] shaped
    ket.

    Parameters
    ----------
    rho : Qobj
         Density matrix

    Returns
    -------
    Qobj
        A Qobj vector containing the vectorized version of the input density matrix

    """
    dim = rho.dims[0][0]
    return qt.Qobj(qt.operator_to_vector(rho), dims = [[dim**2], [1]])


def ness(T, ntsteps, L, f, compute_gradient = False):
    """
    Computes the NESS and, if required, the gradient of the NESS with respect
    to the control parameters of the periodic drivings.

    See :ref:`gradient_floqueto` for details about how the gradient is computed.
    """
    if compute_gradient:
        nu = [g.nu for g in f]
        dgt = [g.gradf for g in f]

    L0 = L.H0.full()
    LVlist = []
    for k in range(len(L.V)):
        LVlist.append([L.V[k].full(), f[k].f])

    nv = len(LVlist)
    if compute_gradient:
        if len(nu) != nv or len(dgt) != nv:
            raise Exception("The lengths of nu and dgt should be equal to the length of LVlist")
        Nu = sum(nu)

    LV = []
    gt = []
    for h in range(nv):
        LV.append(LVlist[h][0])
        gt.append(LVlist[h][1])

    times = np.linspace(0, T, ntsteps+1)
    dt = times[1]

    dim = L0.shape[0]
    d = round(np.sqrt(dim))

    gfunc = np.zeros([nv, ntsteps])
    if compute_gradient:
        gradg = np.zeros([Nu, ntsteps])
    for h in range(nv):
        for j in range(ntsteps):
            gfunc[h, j] = gt[h](times[j], None)
            if compute_gradient:
                l = 0
                for h in range(nv):
                    for k in range(nu[h]):
                        gradg[l, j] = dgt[h](times[j])[k]
                        l = l + 1

    lind = np.zeros([dim, dim, ntsteps], dtype = complex)
    for j in range(ntsteps):
        lind[:, :, j] = L0[:, :]
        for h in range(nv):
            gt_ = gt[h](times[j], None)
            for alpha in range(dim):
                for beta in range(dim):
                    lind[alpha, beta, j] = lind[alpha, beta, j] + gt_ * LV[h][alpha, beta]

    if compute_gradient:
        dlind = np.zeros([Nu, dim, dim, ntsteps], dtype = complex)
        l = 0
        for h in range(nv):
            for k in range(nu[h]):
                for j in range(ntsteps):
                    dgt_ = dgt[h](times[j])[k]
                    dlind[l, :, :, j] = dgt_ * LV[h][:, :]
                l = l + 1
                
    ws = np.zeros(ntsteps, dtype = float)
    for m in range((ntsteps)//2):
        ws[m] = (2.0*np.pi/T) * m
    for m in range((ntsteps)//2, ntsteps):
        ws[m] = (2.0*np.pi/T) * (m-ntsteps)

    lindw = np.zeros([dim, dim, ntsteps], dtype = complex)
    for alpha in range(dim):
        for beta in range(dim):
            lindw[alpha, beta, :] = scp.fft.fft(lind[alpha, beta, :]) / (ntsteps)

    if compute_gradient:
        dlindw = np.zeros([Nu, dim, dim, ntsteps], dtype = complex)
        for k in range(Nu):
            for alpha in range(dim):
                for beta in range(dim):
                    dlindw[k, alpha, beta, :] = scp.fft.fft(dlind[k, alpha, beta, :]) / (ntsteps)

    def tracef(x):
        tr = 0.0
        for j in range(d):
            tr = tr + x[j*(d+1)*ntsteps: j*(d+1)*ntsteps+ntsteps].sum()
        return tr

    fdim = dim * ntsteps
    lindb = cyt.floquetcomponent(dim, ntsteps, ws, lindw)
    if compute_gradient:
        dlindb = cyt.floquetcomponent2(Nu, dim, ntsteps, dlindw)

    b = scp.linalg.null_space(lindb)
    bw = b.reshape([dim, ntsteps])
    bt = np.zeros([dim, ntsteps], dtype = complex)
    for alpha in range(dim):
        bt[alpha, :] = scp.fft.ifft(bw[alpha, :]) * ntsteps
    rho0__ = bt[:, 0]
    rho0 = vectortodm(rho0__)
    trace = rho0.tr()
    trace2 = tracef(b)
    b = b / trace2
    bt = bt / trace2
    rho0 = rho0 / rho0.tr()

    if compute_gradient:

        rhs = np.zeros([dim*ntsteps, Nu], dtype = complex)
        for k in range(Nu):
            rhs[:, k] = - np.matmul(dlindb[k, :, :], b)[:, 0]
        x, residuals, rank, svs = np.linalg.lstsq(lindb, rhs, rcond = None)

        for k in range(Nu):
            x[:, k] = x[:, k] - tracef(x[:, k]) * b[:, 0]

        xw = np.zeros([Nu, dim, ntsteps], dtype = complex)
        for k in range(Nu):
            xw[k, :, :] = x[:, k].reshape([dim, ntsteps])

        xt = np.zeros([Nu, dim, ntsteps], dtype = complex)
        for k in range(Nu):
            for alpha in range(dim):
                xt[k, alpha, :] = scp.fft.ifft(xw[k, alpha, :]) * ntsteps

    if compute_gradient:
        return bt, xt, x
    else:
        return bt


class Nessopt:
    """
    class used for optimization of NESSs
    """
    def __init__(self, target_operator, T, nts, L, glist):
        self.target_operator = target_operator
        self.avector = dmtovector(target_operator).full()[:, 0]
        self.T = T
        self.nts = nts
        self.glist = glist
        self.L = L
        self.L0 = L.H0.full()
        self.LVlist = []
        for k in range(len(L.V)):
            self.LVlist.append([L.V[k].full(), glist[k].f])

    def Afunc(self, y):
        return (1/self.T) * np.vdot(y, self.avector).real

    def F(self, y):
        times = np.linspace(0, self.T, self.nts+1)
        integrand = np.zeros(self.nts+1)
        for j in range(self.nts):
            integrand[j] = self.Afunc(y[:, j])
        integrand[self.nts] = integrand[0]
        return scp.integrate.simps(integrand, times)

    def Fdy(self, dy):
        nu = dy.shape[0]
        nts = dy.shape[2]
        times = np.linspace(0, self.T, nts+1)
        integrand = np.zeros([nu, nts+1])
        for l in range(nu):
            for j in range(nts):
                integrand[l, j] = self.Afunc(dy[l, :, j])
            integrand[l, nts] = self.Afunc(dy[l, :, 0])
        return np.array( [ scp.integrate.simps(integrand[l, :], times) for l in range(nu)] )

    def G(self, u):
        pulses.pulse_collection_set_parameters(self.glist, u)
        y0 = ness(self.T, self.nts, self.L, self.glist)
        return self.F(y0)

    def gradG(self, u):
        pulses.pulse_collection_set_parameters(self.glist, u)
        y0, drhostdu, dy_ = ness(self.T, self.nts, self.L, self.glist, compute_gradient = True)
        Fy0 = self.F(y0)
        gradG = self.Fdy(drhostdu)
        return Fy0, gradG

    def dGdu_fd(self, u, m, delta = 0.01):
        pulses.pulse_collection_set_parameters(self.glist, u)
        def f(x):
            u0 = u.copy()
            u0[m] = x
            return self.G(u0)
        val, err = diff_ridders(f, u[m], delta)
        pulses.pulse_collection_set_parameters(self.glist, u)
        return val

    def gradG_fd(self, u):
        p = u.shape[0]
        dGdunum = np.zeros(p)
        for m in range(p):
            dGdunum[m] = self.dGdu_fd(u, m)
        return dGdunum

    def make_G(self):
        def Gfunction(u, grad = None):
            g1 = self.glist[0]
            g2 = self.glist[1]
            pulses.pulse_collection_set_parameters([g1, g2], u)
            if (grad is not None) and (grad.size > 0):
                Gu, grad[:] = self.gradG(u)
                pulses.pulse_collection_set_parameters([g1, g2], u)
                return Gu
            else:
                return self.G(u)
        return Gfunction



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.



import os
from gitinfo import get_git_info
import pytest

__version__ = "0.0.13"

def isnotebook():
    """Returns True if running within a notebook, False otherwise"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def about():
    """Prints about information about the version

    If available, it also prints out the git hash
    """
    print("Running qocttools version "+__version__)
    gitdir = os.path.dirname(__file__)+'/..'
    git_info = get_git_info(gitdir)
    if git_info is not None:
        print("    Project repository git hash = " + git_info["commit"])
        print("    Project repository last commit date = " + git_info["author_date"])
        print("    Project repository refs = " + git_info["refs"])


def test(nodes = None):
    """Runs the testsuite (the short one)"""
    print("Running qocttools testsuite")
    testsuite_dir = os.path.join(os.path.dirname(__file__), 'tests')
    if nodes == None:
        pytest.main(["-v", os.path.join(testsuite_dir, "short.py")])
    else:
        pytest.main(["-v", "-n {} ".format(nodes), os.path.join(testsuite_dir, "short.py")])



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

import subprocess
import os
import tempfile
import shutil
import numpy as np
from distutils.dir_util import copy_tree
import pytest


def run_test(n, name = None, datafile = None):
    tempdir = tempfile.mkdtemp()
    if type(n) is int:
         n = str(n)
    dirname = os.path.dirname(__file__) + '/../sampleruns/' + n
    copy_tree(dirname, tempdir)
    cdir = os.getcwd()
    os.chdir(tempdir)
    if name is not None:
        fname = name
    else:
        fname = ' sample'+n
    i = subprocess.call('jupyter nbconvert ' + fname + '.ipynb --to python', shell = True)
    j = subprocess.call('python ' + fname +'.py', shell = True)
    datacomp = True
    if datafile is not None:
        data = np.loadtxt(datafile)
        dataref = np.loadtxt(datafile+'.ref')
        datacomp = (data == pytest.approx(dataref))
    os.chdir(cdir)
    if (i, j) == (0 , 0) and datacomp:
        shutil.rmtree(tempdir)
    else:
        print("Failed test directory at " + tempdir)
    return i, j, datacomp


def test_model():
    i, j, datacomp = run_test('model', name = 'model', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_pipulse():
    i, j, datacomp = run_test('pipulse', name = 'pipulse', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_propagators():
    i, j, datacomp = run_test('propagators', name = 'propagators', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_gradient():
    i, j, datacomp = run_test('gradient', name = 'gradient', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_gradient_dissipation():
    i, j, datacomp = run_test('gradient-dissipation', name = 'gradient-dissipation', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_gradient_2perts():
    i, j, datacomp = run_test('gradient-2perts', name = 'gradient-2perts', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_qoctbasic():
    i, j, datacomp = run_test('qoctbasic', name = 'qoctbasic', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_dissipation():
    i, j, datacomp = run_test('dissipation', name = 'dissipation', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_gradient_dissipation_multitarget():
    i, j, datacomp = run_test('gradient-dissipation-multitarget', name = 'gradient-dissipation-multitarget', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_qoctbasic_krotov():
    i, j, datacomp = run_test('qoctbasic-krotov', name = 'qoctbasic-krotov', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

def test_floquet():
    i, j, datacomp = run_test('floquet', name = 'floquet', datafile = 'data')
    assert  (i, j) == (0, 0) and datacomp

# This file should be run as "py.test -v qocttools_tests.py"
# Or, even better: "pytest -n NPROCS --durations=0 -v qocttools_tests.py"
# where NPROCS is the number of cores that can be used in parallel.



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module holds the GdW30 model


"""

import numpy as np
import qutip as qt

def hGdW30(D, E, H):
    """Creates a spin-7/2 Hamiltonian, tailored for the GdW30 molecular nanomagnet.

    Parameters
    ----------
    D : float
       One anysotropy parameter used to construct the Hamiltonian (see notes below).
    E : float
       One anysotropy parameter used to construct the Hamiltonian (see notes below).
    H : ndarray
       A three-dimensional numpy array used to construct the Hamitonian (see notes below).

    Returns
    -------
    Qobj:
        A Qobj operator with the static Hamiltonian modeling the spin system.

    Notes
    -----
    The operator returned follows the formula:

    .. math:: \hat{H}_0 = D(\hat{S}_z^2 - (1/3 S(S+1)) + E(\hat{S}_x^2-\hat{S}_y^2) \
              - g\mu_B \hat{\overrightarrow{S}}\overrightarrow{H}

    """
    S = 7/2
    g = 2.0
    mub = 13996.245 # Bohr magneton, in units of (MHz * h)/tesla
    Sx = qt.jmat(S, "x")
    Sy = qt.jmat(S, "y")
    Sz = qt.jmat(S, "z")
    H0 = D*(Sz**2 - (1/3 * S*(S + 1))) + E*(Sx**2 - Sy**2) - g * mub * (Sx*H[0] + Sy*H[1] + Sz*H[2])
    return H0


def vGdW30(H_m):
    """Creates a perturbation term for the for the GdW30 molecular nanomagnet.

    Parameters
    ----------
    H: : ndarray
       A three-dimensional numpy array used to construct the perturbation term (see notes below).

    Returns
    -------
    Qobj:
        A Qobj operator with the perturbation operator.

    Notes
    -----
    The operator returned follows the formula:

    .. math:: \hat{V} = - g\mu_B \hat{\overrightarrow{S}}\overrightarrow{H}_m

    """
    S = 7/2
    g = 2.0
    mub = 13996.245 # Bohr magneton, in units of (MHz * h)/tesla
    Sx = qt.jmat(S, "x")
    Sy = qt.jmat(S, "y")
    Sz = qt.jmat(S, "z")
    V = -g*mub*(Sx*H_m[0] + Sy*H_m[1] + Sz*H_m[2])
    return V




## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module implements the Krotov's algorithm :cite:p:`Reich2012`.

The current implementation is rather restrictive, and it can only
be applied when:

1. The state is a wavefunction (no density matrices or evolution
   operators).

2. The Hamiltonian is linear in the control field, with only one
   perturbation, i.e. it has the form:

   .. math::
      H(t) = H_0 + f(t) V

3. The pulse is parametrized in real time.

4. The only allowed propagators are RK4 and cfmagnus4.

5. The target has the form:

   .. math::
      F(\\psi(T)) = \\langle \\psi(T)\\vert O  \\vert\\psi(T)\\rangle 
      - \\int_0^T\\!\\!{\\rm d}t\\; \\frac{\\alpha}{S(t)}f^2(t)

   i.e. the objective is to maximize the expectation value of some
   operator :math:`O`, and we have to add a penalty function, depending
   on the parameter :math:`\\alpha > 0` and on the weigth function :math:`S(t)`,
   that can be supplied by the user.


The algorithm that is implemented here is the following:

1. :math:`k = 0`; :math:`f^{(0)}` is the initial guess pulse. Solve:

   .. math::
      i\\frac{\\partial}{\\partial t} \\psi^{(0)}(t) = [H_0 + f^{(0)}V(t)]\\psi^{0}(t)
   .. math::
      \\psi^{(0)}(0) = \\psi_0
   .. math::
      i\\frac{\\partial}{\\partial t} \\chi^{(0)}(t) = [H_0 + f^{(0)}V(t)]\\chi^{0}(t)
   .. math::
      \\chi^{(0)}(T) = O\\psi^{(0)}(T)

2. For :math:`k=1,\dots` until convergence:

   .. math::
      i\\frac{\\partial}{\\partial t} \\psi^{(k)}(t) = 
      [H_0 + \\frac{S(t)}{\\alpha} 
      {\\rm Im} [\\langle \\chi^{(k-1)}\\vert V \\vert\\psi^{(k)}(t)\\rangle ]\\psi^{(k)}(t)
   .. math::
      \\psi^{(k)}(0) = \\psi_0
   .. math::
      f^{(k)}(t) = \\frac{S(t)}{\\alpha} {\\rm Im} 
      [\\langle \\chi^{(k-1)}\\vert V \\vert\\psi^{(k)}(t)\\rangle ]
   .. math::
      i\\frac{\\partial}{\\partial t} \\chi^{(k)}(t) = [H_0 + f^{(k)}V(t)]\\chi^{k}(t)
   .. math::
      \\chi^{(k)}(T) = O\\psi^{(k)}(T)


Convergence is checked by comparing :math:`f^{(k+1)}` with :math:`f^{(k)}`.

"""

import numpy as np
import scipy as sp
import nlopt
import math
from qutip import *
import qocttools.math_extra as math_extra
import qocttools.pulses as pulses
import qocttools.solvers as solvers
from time import time as clocktime


convergence = []

def krotov(solve_method, H, f_initial, psi0, times, O, S,
           alpha = 2.0, 
           tolerance = 1E-03, 
           verbose = False, 
           maxeval = 100, 
           interaction_picture = False):
    """This is the main function on the module.

    It performs the optimization and returns the optimal value of the
    functional, the optimal parameters, and a code signaling the success
    or failure of the process.

    Parameters
    ----------
    solve_method: string
        string indicating the propagation method
    H: hamiltonian
        Hamiltonian
    f_initial: pulse
        pulse to be used as initial guess (on output, it will contain the
        optimized pulse.
    psi0: Qobj
        initial state.
    time: ndarray
        array that contains each time step.
    O: Qobj
        The target operator that we want to optimize.
    S: function
        The penalty function :math:`S(t)`
    alpha: float, default = 2.0
        Penalty factor :math:`alpha`
    tolerance: float, default = 1.0e-3
        Precision for the algorithm
    verbose: bool, default = True
        Whether or not the algorithm will print info to stdout
    maxeval: int, default = 100
        The maximum number of iterations that will be done before the algorithm stops.
    interaction_picture: bool, default = False
        The code will transform the equations to the interaction picture if this is set to True

    Returns
    -------
    ndarray:
        The optimal parameters
    float:
        The optimal target function value
    list:
        A list with data about the convergence history of the algorith
    int:
        An integer flag about the outcome: zero if successful, non-zero otherwise
    """
    global convergence
    H0 = H.H0
    V = H.V
    times_inverted = times[::-1]
    f = []
    for ft in f_initial:
        f.append( check_pulse(ft, times) )
    dim = H0.shape[0]
    obj = psi0.type
    if verbose:  
        print('Object to propagate', obj)
        print('Tolerance', tolerance)
        print('Alpha', alpha)


    # We do iteration zero, which is special.    
    counter = 0
    nprops = 0
    t0 = clocktime()
    psi = solvers.solve(solve_method, H, f, psi0, times, 
                        returnQoutput = False, interaction_picture = interaction_picture)
    psi_final = psi[-1]
    gvalue, pvalue, tvalue = compute_target_function(obj, f, psi_final, times, O, S, alpha, dim)
    chi = coestate_eq(solve_method, H, f, psi_final, times_inverted, O, obj,
                      interaction_picture = interaction_picture)
    t1 = clocktime()
    nprops = nprops + 2
    convergence.append([counter, nprops, tvalue, t1-t0])
    if verbose: write_status(gvalue, pvalue, tvalue, t0, t1, counter)

    old_pulse = f
    counter = 1
    t0 = clocktime()
    psi_final, f = state_eq(solve_method, H0, V, f, chi, psi0, times, S,
                            alpha = alpha,
                            interaction_picture = interaction_picture)
    gvalue, pvalue, tvalue = compute_target_function(obj, f, psi_final, times, O, S, alpha, dim)
    chi = coestate_eq(solve_method, H, f, psi_final, times_inverted, O, obj,
                      interaction_picture = interaction_picture)
    t1 = clocktime()
    nprops = nprops + 2
    convergence.append([counter, nprops, tvalue, t1-t0])
    if verbose: write_status (gvalue, pvalue, tvalue, t0, t1, counter)

    counter=1
    flag=0
    t0 = clocktime()
    while not have_converged(old_pulse, f, times, epsilon = tolerance):
        t0  = clocktime()
        old_pulse = f
        psi_final, f = state_eq(solve_method, H0, V, f, chi, psi0, times, S,
                                alpha = alpha,
                                interaction_picture = interaction_picture)
        gvalue, pvalue, tvalue = compute_target_function(obj, f, psi_final, times, O, S, alpha, dim)
        chi = coestate_eq(solve_method, H, f, psi_final, times_inverted, O, obj,
                          interaction_picture = interaction_picture)
        t1 = clocktime()
        nprops = nprops + 2
        convergence.append([counter, nprops, tvalue, t1-t0])
        if verbose: write_status(gvalue, pvalue, tvalue, t0, t1, counter)
        counter = counter + 1
        if counter+1 >= maxeval:
            print('Maximun number of iterations reached')
            flag=1
            break
    
    if verbose: print("End of optimization")
    gvalue, pvalue, tvalue = compute_target_function(obj, f, psi_final, times, O, S, alpha, dim)    
    #return f.u, gvalue, convergence, flag
    return pulses.pulse_collection_get_parameters(f), gvalue, convergence, flag


def state_eq(solve_method, H0, V, f, chi, psi0, times, S, alpha = 2,
             interaction_picture = False):
    """ This function solves the non-linear Schrödinger equation that is part of
    Krotov's algorithm.
    """
    if solve_method == 'rk4':
        return rk4_nonlinear(H0, V, f, chi, Qobj(psi0), times, S, alpha = alpha, 
                             interaction_picture = interaction_picture)
    elif solve_method == 'cfmagnus4':
        return cfmagnus4_nonlinear(H0, V, f, chi, Qobj(psi0), times, S, alpha = 2,
                                   interaction_picture = interaction_picture)#Function in progress


def compute_target_function(obj, f, psi_final, times, O, S, alpha, dim):
    if obj == 'oper':
        U_target = O.full()
        gvalue = np.matmul(U_target.conj().T, psi_final).trace()
        gvalue = np.absolute(gvalue)**2 / (dim**2)
    else:
        gvalue = abs(expect(O,Qobj(psi_final)))
    #pvalue = - alpha * sp.integrate.simps( f.fu(times[1:-1]) * f.fu(times[1:-1]) / S(times[1:-1]), times[1:-1] )
    pvalue = 0.0
    for ft in f:
        pvalue = pvalue - alpha * sp.integrate.simps( ft.fu(times[1:-1]) * ft.fu(times[1:-1]) / S(times[1:-1]), times[1:-1] )
    tvalue = gvalue + pvalue
    return gvalue, pvalue, tvalue


def write_status(gvalue, pvalue, tvalue, t0, t1, counter):
    print("{:d} gvalue = {:f}\tpvalue = {:f}\ttvalue = {:f} ({:f} s)".format(counter, gvalue, pvalue, tvalue, t1-t0))


def coestate_eq(solve_method, H, f, psiF, times, O, obj, interaction_picture = False):
    """This function calculate the coestate equation with the final condition described by

    .. math::
       \\vert\\chi(T)\\rangle = O\\vert \\Psi(t=T)\\rangle

    In the case of having a state propagating, and

    .. math::                           
       \\vert B(T)\\rangle = Tr [U(T)+*O]*O/d^2

    in the case of having an operator propagating.

    The main idea is to solve it as an linear schrödinger equation but
    *backwards*, so we have to take care of the time and the pulse in order to
    make this correctly (with *realtime* parametrized pulses only is needed to 
    invert the time).
    In principle this should work with any solve_method: rk4, cfm2 or cmf4.
    """
    #Calculating the "initial" coestate (actually the final)
    H0 = H.H0
    v = H.V
    if obj == 'oper':
        dim = H0.shape[0]
        U_target = O.full()
        
        #|B(T)>= Tr[U(T)*U_targ^H]*U_targ/d^2
        ini = U_target * np.matmul(U_target.conj().T, psiF).trace() /dim/dim
        #ini = ini * 2
        ini = Qobj(ini)
    
    else:
        #|Coestate(t=T)>=O|ψ(t=T)>  
        ini = O*Qobj(psiF)
        

    #Solving the equation "backwards"
    coestate=solvers.solve(solve_method, H, f, ini, times,
                           returnQoutput = False,
                           interaction_picture = interaction_picture)

    #Returning the coestate "fordwards", so we can operate with it directly
    return coestate[::-1]


def have_converged (old, new, times, epsilon = 1E-06):
    """ This function checks if the algorithm have converged enough. 

    The idea is to check every time of the pulse.

    Parameters
    ----------
    old:
        The pulse before solving the non-linear schrodinger
    new:
        The pulse after solving the non-linear schrodinger
    epsilon:
        The tolerance we have so we can say the pulse have converged enough

    Returns
    -------
    A boolean which is 'True' if every amplitude have converged, and is 'False' if any
    of the amplitudes have not converged enough
    """
    return False
    for i in range(times.shape[0]):
        if np.abs(old.fu(times[i])-new.fu(times[i]))>epsilon :
            return False
        if new.fu(times[i]) is np.nan:
            raise Exception('Pulse divergence')
        
    return True


def check_pulse (f, times):
    """ This function checks if the pulse parametrization is *realtime*.

    If it is not, it convert the pulse
    into *realtime* parametrization.
    """
    if f.type =='realtime':
        return f
    
    elif f.type =='fourier':
        ut = f.fu(times)
        ft = pulses.pulse("realtime", times[-1], u = ut)
        print('Initial pulse is in Fourier parametrization, changed into Realtime.')
        return ft

    elif f.type =='user_defined':
        ut = f.fu(times)
        ft = pulses.pulse("realtime", times[-1], u = ut)
        print('Initial pulse is in user_defined parametrization, changed into Realtime.')
        return ft

    else:
        raise Exception('The pulse is not well defined')


def rk4_nonlinear(H0, V, f, chi, psi0, times, S, alpha = 2,
                  interaction_picture = False):
    """ This function solves the nonlinear schordinger equation using rk4 as the integrator algorithm.

    It is a modification of the function rk4solver in the solvers module of 
    qocttools, so it can solve a slightly different equation and it can return the new state 
    and the new pulse.
    """
    if type(H0) is not qutip.qobj.Qobj:
        raise TypeError

    #u_new=[]
    u_new = np.zeros((len(f), times.size))
    h0 = H0.full()
    #v = V.full()
    v = [x.full() for x in V]
    dt = times[1]-times[0]
    dim = H0.shape[0]

    #What are we propagating?
    if psi0.type == 'oper':
        psi = psi0.full()
        y = np.zeros((2, dim, dim), dtype = complex)
        y[0] = psi
        y[1] = chi[0].copy()
        obj = 'oper'
    else:
        psi = psi0.full()[:, 0]
        y = np.zeros((2, dim), dtype = complex)
        y[0] = psi
        y[1] = chi[0].copy()
        obj = 'ket'


    def nlf(t, y):
        psi = y[0]
        chi = y[1]
        fy = y.copy()
        if interaction_picture:
            htpsi = np.zeros_like(h0)
            htchi = np.zeros_like(h0)
            for k in range(len(f)):
                vi = solvers.intoper(v[k], np.diag(h0), t)
                if obj == 'ket' :
                    V_Matrix_Element = np.vdot(chi, np.matmul(vi, psi))
                elif obj == 'oper':
                    V_Matrix_Element = np.matmul(chi.conj().T, np.matmul(vi, psi)).trace()
                htpsi = htpsi + (S(t) * V_Matrix_Element.imag / alpha) * vi
                htchi = htchi + f[k].fu(t) * vi
            fpsi = - 1j * np.matmul(htpsi, psi)
            fchi = - 1j * np.matmul(htchi, chi)
        else:
            htpsi = h0.copy()
            htchi = h0.copy()
            #for vi in v:
            for k in range(len(f)):
                if obj == 'ket' :
                    V_Matrix_Element = np.vdot(chi, np.matmul(v[k], psi))
                elif obj == 'oper':
                    V_Matrix_Element = np.matmul(chi.conj().T, np.matmul(v[k], psi)).trace()
                htpsi = htpsi + (S(t) * V_Matrix_Element.imag / alpha) * v[k]
                htchi = htchi + f[k].fu(t) * v[k]
            fpsi = - 1j * np.matmul(htpsi, psi)
            fchi = - 1j * np.matmul(htchi, chi)

        fy[0] = fpsi
        fy[1] = fchi
        return fy

    # u_new.append(S(0) * V_Matrix_Element.imag / alpha)
    for k in range(len(f)):
        if obj == 'ket' :
            V_Matrix_Element = np.vdot(chi[0], np.matmul(v[k], psi))
        elif obj == 'oper':
            V_Matrix_Element = np.matmul(chi[0].conj().T, np.matmul(v[k], psi)).trace()
        u_new[k, 0] = S(0) * V_Matrix_Element.imag / alpha

    for j in range(times.size-1):
        y = math_extra.rk4(y, nlf, times[j], dt)

        psi_ = y[0]
        chi_ = y[1]

        for k in range(len(f)):
            if interaction_picture:
                vi = solvers.intoper(v[k], np.diag(h0), times[j+1])
            else:
                vi = v[k].copy()
            if obj == 'ket' :
                V_Matrix_Element = np.vdot(chi_, np.matmul(vi, psi_))
            elif obj == 'oper':
                V_Matrix_Element = np.matmul(chi_.conj().T, np.matmul(vi, psi_)).trace()
            u_new[k, j+1] = S(times[j+1]) * V_Matrix_Element.imag / alpha
    
    #Creating the new pulse
    #f_new = pulses.pulse("realtime", times[-1], u = np.array(u_new))
    f_new = []
    for k in range(len(f)):
        f_new.append( pulses.pulse("realtime", times[-1], u = u_new[k, :]) )

    return psi_, f_new

    
def cfmagnus4_nonlinear(H0, V, f, chi, psi0, time, S, alpha = 2,
                        interaction_picture = False, cops = None):
    """ This function solves the nonlinear schordinger equation using cfmagnus4 as the integrator algorithm.
    """
    if type(H0) is not qutip.qobj.Qobj:
        raise TypeError

    u_new = []
    h0 = H0.full()
    dt = time[1]-time[0]
    dim = H0.shape[0]

    if psi0.type == 'oper':
        psi = psi0.full()
        obj = 'oper'
    else:
        psi = psi0.full()[:, 0]
        obj = 'ket'

    a1 = (3.0-2.0*np.sqrt(3.0))/12.0
    a2 = (3.0+2.0*np.sqrt(3.0))/12.0
    c1 = 0.5 - np.sqrt(3.0)/6.0
    c2 = 0.5 + np.sqrt(3.0)/6.0

    if obj == 'ket':
        V_Matrix_Element = np.vdot(chi[0], np.matmul(V.full(), psi))
   
    else:
        V_Matrix_Element = np.matmul(chi[0].conjugate().T, np.matmul(V.full(), psi)).trace()

    u_new.append(S(0) * V_Matrix_Element.imag / alpha)
    for j in range(time.size-1):
        t1 = time[j] + c1*dt
        t2 = time[j] + c2*dt
        if interaction_picture:
            # We will assume that H0 is diagonal on entry.
            M1 = np.zeros_like(h0)
            M2 = np.zeros_like(h0)
            #for k in range(len(V)):
            vi1 = solvers.intoper(V.full(), np.diag(h0), t1)
            vi2 = solvers.intoper(V.full(), np.diag(h0), t2)
            
            if obj == 'ket':
                V_Matrix_Element_1 = np.vdot(chi[j], np.matmul(vi1, psi))
                V_Matrix_Element_2 = np.vdot(chi[j], np.matmul(vi2, psi))

            else:
                V_Matrix_Element_1 = np.matmul(chi[j].conjugate().T, np.matmul(vi1, psi)).trace()
                V_Matrix_Element_2 = np.matmul(chi[j].conjugate().T, np.matmul(vi2, psi)).trace()

            #u_new.append(S(time[t])*)

            M1 = M1 + a1 * S(t1) * V_Matrix_Element_1.imag / alpha * vi1 + a2 * S(t2) * V_Matrix_Element_2.imag / alpha * vi2
            M2 = M2 + a2 * S(t1) * V_Matrix_Element_1.imag / alpha * vi1 + a1 * S(t2) * V_Matrix_Element_2.imag / alpha * vi2


        else:
            M1 = (a1 + a2) * h0
            M2 = (a1 + a2) * h0
            #for k in range(len(V)):
            
            M1 = M1 + a1 * S(t1) * V_Matrix_Element.imag / alpha * V.full() + a2 * S(t2) * V_Matrix_Element.imag / alpha * V.full()
            M2 = M2 + a2 * S(t1) * V_Matrix_Element.imag / alpha * V.full() + a1 * S(t2) * V_Matrix_Element.imag / alpha * V.full()

        M = M2
        psi = solvers.exppsi(2*M, cops, dt/2, psi)
        M = M1
        psi = solvers.exppsi(2*M, cops, dt/2, psi)

        #Saving Values
        if obj == 'ket':
            V_Matrix_Element = np.vdot(chi[j+1], np.matmul(V.full(), psi))
            
        else:
            V_Matrix_Element = np.matmul(chi[j+1].conjugate().T, np.matmul(V.full(), psi)).trace()
        
        u_new.append(S(time[j+1]) *V_Matrix_Element.imag / alpha)



    #Creating the new pulse
    f_new=pulses.pulse("realtime", time[-1], u = np.array(u_new))
    
    return psi, f_new


def coestate_interpolator(chi, times, coestate_flag, solve_method):
    """ This function interpolates the value of the coestate.

    In this way, we can solve properly the non-linear Schrodinger equation with the rk4 and cfmagnus4 integrators.
    This function returns the value of the coestate in the time that´s required
    """
    #We separate coestate_flag into int number (i), and decimal number (flag)
    flag, i = math.modf(coestate_flag)
    i=int(i)
    #If coestate_flag is an int number it´s not needed to interpolate
    if flag == 0:
        return chi[i]
    
    else:
        #Lets see if it´s solve by rk4 or cfmagnus4
        if solve_method == 'rk4':
            #If it´s rk4 and coestate_flag is not an int, neccesary we are in the case were we need to get the coestate in t+dt*0,5
            t_obj=times[i]+(times[i+1]-times[i])/2
            return chi[i]+(chi[i+1]-chi[i])/(times[i+1]-times[i])*(t_obj-times[i])

        elif solve_method == 'cfmagnus4':  
            return
        else:
            raise Exception('Don´t know propagation method')



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

""" This module includes the 'pulse' class, that is used to contain
the time-dependent real function that is used to define the perturbation
part of the Hamiltonian.

Also, it contains definitions of 'typical' pulse forms, such as the
pi-pulse.
"""

import numpy as np
import scipy as sp
import qocttools.cythonfuncs as cyt
import qutip as qt



class pulse:
    """Definition of a class to hold control functions, a.k.a. pulses.

    The "pulses" are real time-dependent functions defined in the interval
    :math:`[0, T]`, defined as functions :math:`f(u_1, u_2, ..., u_P, T)`, where 
    :math:`u = u_1, ..., u_P` are parameters (the *control parameters*). These pulses
    are the control functions used by qocttools.

    Depending on the `type` of the pulse, the parametrized form of these pulses is:

    1. fourier

       In this case, the pulse is a simple Fourier expansion:

       .. math::
          f(u, t) = \\frac{u_0}{\sqrt{T}} + \\frac{2}{\\sqrt{T}}
          \\sum_{i=1}^M \\left[ u_{2i}\\cos(\\omega_i t) + u_{2i-1}\\sin(\\omega_i t)\\right]

       The pulse therfore requires an odd number of parameters, :math:`P = 2M + 1`. The
       frequencies are :math:`\omega_i = i\omega_0 = i\\frac{2\\pi}{T}` for :math:`i=1, \\dots, M`.
       The cutoff M will be decided by the number of parameters that are passed when
       initializing the object. Therefore, in order to initialize a Fourier pulse with :math:`M=5`:

       .. code-block:: python

          M = 5
          u = np.zeros(2*M+1)
          f = pulses.pulse('fourier', T, u)

    2. bound_fourier

       This is a modification of the normal Fourier expansion, that ensures that the absolute value
       of the pulse is never larger than a certain bound. The definition is:

       .. math::
          f(u, t) = \\Phi(g(u, t))

       where :math:`g(u, t)` is a normal Fourier expansion as the one given above, and:

       .. math::
          \\Phi(x) = \\frac{\\kappa x}{\\kappa+\\vert x\\vert}

       The value of :math:`\kappa` must be supplied by the `bound` argument:

       .. code-block:: python

          M = 5
          kappa = 3.0
          u = np.zeros(2*M+1)
          f = pulses.pulse('fourier', T, u, bound = kappa)


    3. realtime

       In this case, the pulse is given by a piecewise-constant function. If the time
       discretization is given by :math:`t_0 = 0, t_1 = \Delta t, t_2 = 2\Delta t,\dots,t_{N-1}=(N-1)\Delta t = T`,
       then the function is given by:

       .. math::
          f(u, t) = \\sum_{i=0}^{N-1} u_i {\\bf 1}_{[t_i,t_{i+1}]}

       where :math:`{\\bf 1}_{[t_i,t_{i+1}]}` is the indicator function in the interval
       :math:`[t_i,t_{i+1}]`

       The number :math:`N` is given by the number of elements in the numpy array `u` used
       when creating th epulse, and therefore :math:`\Delta t` is computed accordingly, given :math:`T`.

       In fact, although still experimental and therefore undocumented, one can also use some interpolating
       function to obtain the values in the interior of the intervals, so that the function is smooth


    4. enveloped_fourier

       This is a normal Fourier expansion, but multiplied by a function :math:`S(t)`:

       .. math::
          f(u, t) = S(t) g_{\\rm Fourier}(u, t)

       The function :math:`S(t)` must be passed to the function :meth:`assign_envelope_function`,
       and is simply a function of time:

       .. code-block:: python

          def S(t):
              return np.sin(t)
          f = pulses.pulse('enveloped_fourier', T, u, bound = kappa)
          f.assign_envelope_function(S)

    5. user_defined

       This is the most general possibility: the user defines the pulse as a
       a python function. The function must be communicated to the pulse object
       using the :meth:`assign_user_defined_function`. One should also define
       a function with the derivatives of the function with respect to the parameters.
       Here is an example using only three of parameters:

       .. math::

          f(u, t) = u_0 \\cos(u_1 t+ u_2)

       .. code-block:: python

          def fu(t, u):
              return u[0] * np.cos(u[1]*t + u[2])
          def dfdu(t, u, m):
              # m is an integer that signals what is the parameter with respect to which the 
              # derivative is taken
              if m == 0:
                  return np.cos(u[1]*t + u[2])
              elif m == 1:
                  return -u[0] * t * np.sin(u[1]*t + u[2])
              elif m == 2:
                  return -u[0] * np.sin(u[1]*t + u[2])
          u = np.array([1.0, 0.5, 0.0])
          f = pulses.pulse('user_defined', T, u)
          f.assign_user_defined_function(fu, dfdu)

       If one is going to use a gradient-less QOCT algorithm, there is no need to define
       the derivative function, and one can just pass the None value.


    Parameters
    ----------
    type : string
        The type of function: 'fourier', 'bound_fourier', 'realtime', 'user_defined', or
        'enveloped' fourier
    T : float
        The duration of the pulse
    u : ndarray
        The parameters
    bound : float, default = None
        For the 'bound_fourier' type, the bound


    Attributes
    ----------
    type : string
    T : float
    u : ndarray
    nu : int
    bound : float

    """
    def __init__(self, type, T, u, bound = None):
        self.type = type
        self.T = T
        self.set_parameters(u)
        self.nu = u.shape[0]
        self.constraints = []
        self.bound = bound

    def print(self, filename):
        """ Prints a pulse to a file called 'filename'.

        It does not print the possible constraint functions.

        Parameters
        ----------
        filename : str
            Name of the file where the info about the pulse is written.
        """
        f = open(filename, 'w')
        f.write(self.type+'\n')
        f.write('{}\n'.format(self.T))
        f.write('{}\n'.format(self.nu))
        for j in range(self.nu):
            f.write(' {} '.format(self.u[j]))
        f.write('\n')
        f.write('{}\n'.format(self.bound))
        f.write('{}\n'.format(len(self.constraints)))
        f.close()

    def set_constraint(self, which, val = 0.0):
        """ Sets a constraint on the shape of the pulse

        """
        if which == 'zero_boundaries':
            def g(u, grad):
                M = len(u) // 2
                if grad.size > 0:
                    gu = 0.5 * u[0]
                    grad[:] = 0.0
                    grad[0] = 0.5
                    for k in range(1, M+1):
                        gu = gu + u[2*k]
                        grad[2*k] = 1.0
                else:
                    gu = 0.0
                    for k in range(1, M+1):
                        gu = gu + u[2*k]
                return gu
            self.constraints.append(g)
        elif which == 'boundaries_values':
            def g(u, grad):
                M = len(u) // 2
                if grad.size > 0:
                    gu = u[0]
                    grad[:] = 0.0
                    grad[0] = 1.0
                    for k in range(1, M+1):
                        gu = gu + 2.0 * u[2*k]
                        grad[2*k] = 2.0
                else:
                    gu = u[0]
                    for k in range(1, M+1):
                        gu = gu + 2.0 * u[2*k]
                return gu - val * np.sqrt(self.T)
            self.constraints.append(g)
        elif which == 'zero_average':
            def g(u, grad):
                if grad.size > 0:
                    gu = u[0]
                    grad[0] = 1.0
                    grad[1:] = 0.0
                else:
                    gu = u[0]
                return gu
            self.constraints.append(g)
        else:
            print("Unknown constraint: {}.".format(which))
        return None

    def set_parameters(self, u):
        """Sets the parameters :math:`u = u_1, \dots, u_P` on the pulse.

        """
        self.u = u.copy()
        self.nu = u.shape[0]
        if self.type == 'realtime':
            times = np.linspace(0.0, self.T, u.shape[0])
            self.interpolator = sp.interpolate.interp1d(times, u, fill_value = 'extrapolate')

    def f(self, t, args):
        """The value of the pulse at time t

        """
        return self.fu(t)

    def assign_user_defined_function(self, fu, dfu):
        """For user-defined functions, set the function definition, and the definition of the gradient of the function.

        """
        self.user_defined_function = fu
        self.user_defined_dfunction = dfu


    def assign_envelope_function(self, efunc):
        """Sets the envelope function that may be multiplied by the pulse itself.

        """
        self.user_envelope_function = efunc


    def envelope(self, t):
        """Returns the value of the envelope function at time t

        """
        if self.type == 'enveloped_fourier' :
            if isinstance(t, float):
                return cyt.fourierexpansion(t, self.T, self.nu, self.u)
            else:
                return f_FE(t, self.T, self.u)
        else:
            return None


    def fu(self, t, u = None):
        """Returns the value of pulse at time t

        Optionally, one can pass the control parameters u and in that way those
        are updated before the computation.
        """
        if isinstance(t, int):
            t = float(t)
        if u is not None:
            self.u = u.copy()
        if self.type == 'fourier' :
            return cyt.fourierexpansion(t, self.T, self.nu, self.u)
        elif self.type == 'realtime' :
            return self.interpolator(t)
        elif self.type == 'user_defined' :
            return self.user_defined_function(t, self.u)
        elif self.type == 'enveloped_fourier' :
            return cyt.fourierexpansion(t, self.T, self.nu, self.u) * self.user_envelope_function(t)
        elif self.type == 'bound_fourier' :
            ft = cyt.fourierexpansion(t, self.T, self.nu, self.u)
            return self.bound * ft / (self.bound + np.abs(ft))


    def gradf(self, t):
        """Grdient of the function at time t

        "Gradient of the function" means the gradient with respect to all
        the control parameters.
        """
        res = np.zeros(self.nu)
        if self.type == 'fourier':
            res = cyt.gradfourierexpansion(t, self.T, self.nu, self.u)
        elif self.type == 'realtime':
            for m in range(self.nu):
                res[m] = f_realtime_der(t, self.T, self.u, m)
        elif self.type == 'user_defined' :
            for m in range(self.nu):
                res[m] = self.user_defined_dfunction(t, self.u, m)
        elif self.type == 'enveloped_fourier' :
            res = cyt.dfourierexpansion(t, self.T, self.nu, self.u) * self.user_envelope_function(t)
        elif self.type == 'bound_fourier' :
            ft = cyt.fourierexpansion(t, self.T, self.nu, self.u)
            dft = cyt.gradfourierexpansion(t, self.T, self.nu, self.u)
            res = self.bound**2 * dft / ( self.bound + np.abs(ft) )**2
        return res


    def dfu(self, t, m, u = None):
        """Derivative of the pulse with respect to the m-th parameter, at time t

        """
        if u is not None:
            self.u = u.copy()
        if self.type == 'fourier' :
            return cyt.dfourierexpansion(t, self.T, self.nu, self.u, m)
        elif self.type == 'realtime' :
            return f_realtime_der(t, self.T, self.u, m)
        elif self.type == 'user_defined' :
            return self.user_defined_dfunction(t, self.u, m)
        elif self.type == 'enveloped_fourier' :
            return cyt.dfourierexpansion(t, self.T, self.nu, self.u, m) * self.user_envelope_function(t)
        elif self.type == 'bound_fourier' :
            dft = cyt.dfourierexpansion(t, self.T, self.nu, self.u, m)
            ft = cyt.fourierexpansion(t, self.T, self.nu, self.u)
            return self.bound**2 * dft / ( self.bound + np.abs(ft) )**2


    def fitparams(self, fref, times, u0):
        """Fit the parameters of a function to the best mpossible much wrt a reference function

        """
        def fitfunction(t, *u):
            return self.fu(t, np.array(u))
        optu, cov = sp.optimize.curve_fit(fitfunction, times, fref, p0 = u0)
        return optu


    def fw(self, t, u, w = None, explicit_formula = False):
        """Fourier transform of the pulse

        """
        if w is not None:
            if not explicit_formula:
                fw = np.zeros(w.size, dtype = complex)
                for j in range(w.size):
                    omega = w[j]
                    for i in range(t.size-1):
                        fw[j] = fw[j] + np.exp(-1j * t[i] * w[j]) * self.fu(t[i])
                return fw * (t[1]-t[0]) / (np.sqrt(2*np.pi))
            else:
                def Piw(w, T):
                    return (T * np.exp(1j*w*T/2) / np.sqrt(2*np.pi)) * np.sinc((w*T/2.0)/np.pi)
                nfreqs = w.size
                M = int((u.size-1)/2)
                T = self.T
                fw = np.zeros(w.size, dtype = complex)
                freqs = np.zeros(M+1)
                for j in range(1, M+1):
                    freqs[j] = (2.0*np.pi/T)*j
                for i in range(nfreqs):
                    fw[i] = Piw(w[i], T) * u[0]
                    for j in range(1, M+1):
                        fw[i] = fw[i] \
                                -1j * 0.5 * (Piw(w[i]-freqs[j], T) - Piw(w[i]+freqs[j], T)) * 2 * u[2*j-1]
                        fw[i] = fw[i] \
                                + 0.5 * (Piw(w[i]-freqs[j], T) + Piw(w[i]+freqs[j], T)) * 2 * u[2*j]
                return fw / np.sqrt(T)
        else:
            fw_ = sp.fft.fft(self.fu(t[:-1])) * (t[1]-t[0])/(np.sqrt(2.0*np.pi))
            w = np.zeros_like(t[:-1])
            for j in range(w.size):
                w[j] = (2.0*np.pi/t[-1])*j
            return fw_, w


def read_pulse(filename):
    """ Reads a pulse from the info contained in file.

    Parameters
    ----------
    filename : str
        Name of the file where the info about the pulse is written.

    Returns
    -------
    pulse
        The pulse whose information was previously written to filename.
    """

    f = open(filename, 'r')
    lines = f.read().splitlines()
    pulse_type = str(lines[0])
    T = float(lines[1])
    nu = int(lines[2])
    u = np.zeros(nu)
    for j in range(nu):
        u[j] = lines[3].split()[j]
    try:
        bound = float(lines[4])
    except:
        bound = None
    fpulse = pulse(pulse_type, T, u, bound = bound)
    f.close()
    return fpulse


def pulse_collection_parameter_range(f, j):
    """Returns the 'parameter range' of a given pulse in a collection

    Given a list of pulses 'f', the full control parameter list will
    be an array joining all the control parameters of each of of them.
    This function returns the indexes that would correspond, in that
    array, to one of the pulses.

    Thus, for example, if we have two pulses with nu1 and nu2 parameters
    each, the range of the first one would be (0, nu1), whereas the range
    of the second would be (nu1, nu1+nu2).

    Parameters
    ----------
    f : list of pulse
        A list with pulse objects
    j : int
        The pulse for which we want to get the parameter range.

    Returns
    -------
    list of int
        a list of two integer numbers, with the starting index of the parameters
        corresponding to the first pulse, and the final index.
    """
    k = 0
    if j > 0:
        for n in range(j):
            k = k + f[j].nu
    l = k + f[j].nu
    return [k, l]


def pulse_collection_set_parameters(f, u):
    """Sets the parameters of a collection of pulses

    It receives a numpy array of parameters u, whose dimension should
    be equal to the summ of parameters of the pulses in the collection f.

    """
    k = 0
    for j in range(len(f)):
        f[j].set_parameters(u[k:k+f[j].nu])
        k = k + f[j].nu


def pulse_collection_get_parameters(f):
    """Sets the parameters of a collection of pulses

    It receives a numpy array of parameters u, whose dimension should
    be equal to the summ of parameters of the pulses in the collection f.

    """
    nutotal = 0
    for j in range(len(f)):
        nutotal = nutotal + f[j].nu
    u = np.zeros(nutotal)
    k = 0
    for j in range(len(f)):
        u[k:k+f[j].nu] = f[j].u[:]
        k = k + f[j].nu
    return u


def pulse_collection_l(f, m):
    n = len(f)
    k = 0
    for i in range(n):
        k = k + f[i].nu
        if m < k:
            return i

def pulse_collection_j(f, m):
    n = len(f)
    l = pulse_collection_l(f, m)
    if l is not None:
        nprevious = 0
        for k in range(l):
            nprevious = nprevious + f[k].nu
            #print("nprevious = {}".format(nprevious))
        return m - nprevious


def pulse_constraint_functions(f, param_range):
    """Given a pulse, returns a list with all its constraint functions.

    The constraint functions contained in the pulse object are functions
    of the nu control parameters on which the pulse depends. In contrast,
    the functions returned by this function are functions of all the
    control parameters in a list of pulses. Therefore, we need
    as an input a 'parameter range', i.e. the starting and finishin indexes
    of the parameters of this particular pulse in the list.

    """
    g = []
    for cnstr in f.constraints:
        def make_f(cnstr):
            def g(u, grad):
                if grad.size > 0:
                    grad[:] = 0
                k = param_range[0]
                l = param_range[1]
                uj = u[k:l]
                gu = cnstr(uj, grad[k:l])
                return gu
            return g
        g.append(make_f(cnstr))
    return g


def pi_pulse(t, u):
    """A pi-pulse value at time t, parameters specifiey by the ndarray u

    Returns the pi-pulse value at a specific time
    
    Input
        A: pi-pulse amplitude
        w: pi-pulse frecuency
        t0: time in which start the propagation of the pulse
        t: time in which the pulse is calculated
        t_duration: pi-pulse duration
    Output:
        if the t value in which the pulse calculation is requested
        is between t0 and t0 + t_duration, return the value of a pi-pulse
        with frecuency w and amplitude A, otherwise return 0.
    """
    A = u[0]
    w = u[1]
    t0 = u[2]
    t_duration = u[3]
    phi = u[4]
    if isinstance(t, np.ndarray):
        ft = np.zeros_like(t)
        for j in range(t.shape[0]):
            if t[j] >= t0 and t[j] <= (t0 + t_duration):
                ft[j] = A * np.cos(w*(t[j]-t0)+phi)
            else:
                ft[j] = 0.0
        return ft
    else:
        if t >= t0 and t <= (t0 + t_duration):
            return A*np.cos(w*(t-t0)+phi)
        else:
            return 0.0

def pi_pulse_chain(t, u):
    """
    Returns the value of a chain of pi pulse at a specific time
    """
    A = u[0]
    n = int(u[1])
    w = np.zeros(n)
    length = np.zeros(n)
    for i in range(n):
        w[i] = u[2+i]
        length[i] = u[2+n+i]

    y = 0.0
    t_acumulation = 0.0
    for i in range(n):
        y += pi_pulse(t, np.array([A, w[i], t_acumulation, length[i], 0.0]))
        t_acumulation += length[i]
    return y


def f_realtime(t, T, u):
    M = u.shape[0]
    deltat = T / (M-1)
    i = (np.round(t/deltat)).astype(int)
    return u[i]


def f_realtime_der(t, T, u, m):
    if isinstance(t, np.ndarray):
        ntimes = t.shape[0]
        M = u.shape[0]
        deltat = T / (M-1)
        dfdu = np.zeros(ntimes)
        for k in range(ntimes):
            i = (np.round(t[k]/deltat)).astype(int)
            if m == i:
                dfdu[k] = 1.0
        return dfdu
    else:
        M = u.shape[0]
        deltat = T / (M-1)
        i = (np.round(t/deltat)).astype(int)
        if m == i:
            return 1.0
        else:
            return 0.0


def rotation(theta, n, dim = 2, i = None, j = None):
    """Returns a rotation operator between two levels, given an angle an axis.

    The rotation affects level i and j (or the only two levels, if dim = 2, which
    is the default). The angle of rotation is given by theta, and the axis is
    given by the unitary vector n.

    Parameters
    ----------
    theta : float
        The rotation angle.
    n : ndarray
        A three-dimensional float array containing a unit vector.
    dim : int, default = 2
        The dimension of the rotation operator that will be created.
    i : int, default = None
        One of the states affected by the rotation.
    j : int, default = None
        One of the states affected by the rotation.

    Returns
    -------
    Qobj:
        A Qobj operator with the rotation operator.

    """
    rot_ = (-1j * (theta/2) * (n[0] * qt.sigmax() + n[1] * qt.sigmay() + n[2] * qt.sigmaz()) ).expm()
    if dim > 2:
        rotmatrix = np.eye(dim, dtype = complex)
        rotmatrix[i, i] = rot_.full()[0, 0]
        rotmatrix[i, j] = rot_.full()[0, 1]
        rotmatrix[j, i] = rot_.full()[1, 0]
        rotmatrix[j, j] = rot_.full()[1, 1]
        return qt.Qobj(rotmatrix)
    else:
        return rot_


def rotationpulse(axis, A, omega, mu0, theta, t0 = None):
    """Returns a pulse that implements a rotation betwen two states.

    Given an amplitude for a pulse A, and two levels characterized
    by an energy difference omega and a coupling mu0, creates a pulse
    object that implements a rotation between those two levels
    in direction given by axis (0 => x, 1 => y, 2 => z), and angle 
    theta.

    Parameters
    ----------
    axis : int
        The rotation index (0 => x, 1 => y, 2 => z).
    A : float
        The amplitude of the pulse.
    omega : float
        The frequency difference between the levels.
    mu0 : float
        The coupling matrix element
    theta : float
        The rotation angle

    Returns
    -------
    pulse:
        A pulse object with the pulse that produces the
        rotation.

    """
    argmu0 = np.angle(mu0)
    if axis == 1:
        if theta > 0:
            phi = -np.pi/2 - argmu0
        else:
            phi = np.pi/2 - argmu0
        if t0 is not None: phi = phi + omega * t0
    elif axis == 0:
        if theta > 0:
            phi = - argmu0
        else:
            phi = -argmu0 + np.pi
        if t0 is not None: phi = phi + omega * t0
    elif axis == 2:
        res = []
        if t0 is None:
            res.append(rotationpulse(0, A, omega, mu0, -np.pi/2)[0])
            res.append(rotationpulse(1, A, omega, mu0, theta)[0])
            res.append(rotationpulse(0, A, omega, mu0, np.pi/2)[0])
        else:
            res.append(rotationpulse(0, A, omega, mu0, -np.pi/2, t0 = t0)[0])
            res.append(rotationpulse(1, A, omega, mu0, theta, t0 = res[-1].T)[0])
            res.append(rotationpulse(0, A, omega, mu0, np.pi/2, t0 = res[-1].T)[0])
        return res
    duration = np.abs(theta) / (A * np.abs(mu0))
    if t0 is None:
        T = duration
        u = np.array([A, omega, 0.0, duration, phi])
    else:
        T = duration + t0
        u = np.array([A, omega, t0, duration, phi])
    ft = pulse('user_defined', T, u)
    ft.assign_user_defined_function(pi_pulse, None)
    return [ft]






## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module contains the Target class, and associated procedures.

"""

import jax; jax.config.update('jax_platform_name', 'cpu')
import jax.numpy as jnp
from jax import grad
import numpy as np
import qocttools.math_extra as math_extra
import qocttools.floquet as floquet
import qutip as qt


class Target:
    """The class that holds the definition of the target.

    The user can choose between (1) writing a Python function that defines
    the target functional (in general, it must be accompanied of a
    Python function that computes the functional derivative with respect
    to the wavefunction, evolution operator or density matrix), and (2)
    using one of the predefined target types defined by the code. The
    former is chosen by setting the targettype to "generic", whereas
    for the latter one must use either "expectationvalue", "evolutionoperator",
    or "floquet".

    Parameters
    ----------
    targettype: string
        The code admits the following types of targets:

        1. generic

           The user may input any function as target, defined by
           the driver program, using the `Fyu`, `dFdy` and `dFdu` parameters. See below
           for information about the proper definition of these functions.

        2. expectationvalue

           The target is the expectation value of some operator:

           .. math::

              F(\\Psi, u) = \\langle \\Psi(T)\\vert O \\vert \\Psi(T)\\rangle

           or, if the multitarget option is used,

           .. math::

              F(\\Psi, u) = \\frac{1}{Q} \\sum_i \\langle \\Psi_i(T)\\vert O_i \\vert \\Psi_i(T)\\rangle

           Here, :math:`Q` is the number of targets. The operator(s) should be provided 
           using the `operator` argument (below). If one is using density matrices 
           instead of wave functions to describe open systems, then:

           .. math::

              F(\\rho, u) = {\\rm Tr}{\\rho(T)O}

           or, if the multitarget option is used,

           .. math::

              F(\\rho_i, u) = \\frac{1}{Q} \\sum_i {\\rm Tr}{\\rho_i(T)O_i}

           This option is simple to use, one just needs to instantiate a Target class object as:

           .. code-block:: python

              tg = target.Target('expectationvalue', operator = O)

           where "O" is a qutip object containing the operator whose expectation value is to be
           maximized. It must be a list of operators, if optimizing in multitarget mode.

        3. evolutionoperator

           The goal now is to find a set of control parameters that leads to an evolution of
           the system characterized by a fixed target evolution operator. This target evolution
           operator (or operators, if a multitarget setup is used), is provided by the argument
           `Utarget`. The target definition is:

           .. math::

              F(U, u) = \\frac{1}{Q} \\sum_i \\vert U(T) \\cdot U^{(i)}_{\\rm target} \\vert^2

           In this case, the creation of the target object would be:

           .. code-block:: python

              tg = target.Target('evolutionoperator', Utarget = U)

           where U would be the (list of) unitary operators.

        4. floquet

           (undocumented)

    Fyu: function

        If `targettype == "generic"`, the user should pass the function used to 
        define the target. The interface should be:

        .. code-block:: python

           def Fyu(y: qutip.Qobj or list of qutip.Qobj, u: ndarray) -> float:

        The function `Fyu` should receive as argument a system state (be it a Hilbert
        state, evolution operator, or density matrix), that is supposed to be the state
        at the end of the propagation, and a control parameters set. It should output
        the target functional value. It may also take as argument a list of states, if
        one is running the optimizaion in *multitarget* mode.

    dFdy: function

        If `targettype == "generic"`, the user should pass the function used to 
        define the target (`Fyu`), **and** the derivative of that function with respect to
        the state, which would be given by this `dFdy` parameter. The interface should be:

        .. code-block:: python

           def dFdy(y: qutip.Qobj or list of qutip.Qobj, u: ndarray) -> list of float:

    dFdu: function

        If `targettype == "generic"`, the user should pass the function used to 
        define the target (`Fyu`), **and, if it is not zero**, the derivative of that function with respect to
        the control parameters, which would be given by this `dFdu` parameter. The interface should be:

        .. code-block:: python

           def dFdu(u: ndarray, m: int) -> float:

    operator: qutip.Qobj or list of qutip.Qobj

        If `targettype == "expectationvalue"`, this argument should be the operator whose expectation
        value is to be maximized. In multitarget mode, this would be a list of operators, one for each
        system.

    Utarget: qutip.Qobj or list of qutip.Qobj

        If `targettype == "evolutionoperator"`, this argument should be the unitary operator
        that is set as target. In multitarget mode, this would be a list of unitaries, one for each
        system.


    Examples
    ----------

    Let us show a simple example of the use of the "generic" target type. The following would
    be equivalent to using the "expectationvalue" targe type with operator O:

    .. code-block:: python

       def Fpsi(psi, u):
           return qt.expect(O, psi)
       def dFdpsi(psi, u):
           return O * psi
       tg = target.Target("generic", Fyu = Fpsi, dFdy = dFdpsi)



    """
    def __init__(self, targettype, 
                 Fyu = None,
                 dFdy = None,
                 dFdu = None,
                 operator = None,
                 Utarget = None,
                 alpha = None,
                 S = None,
                 targeteps = None,
                 T = None,
                 fepsilon = None,
                 dfdepsilon = None):
        self.targettype = targettype
        if targettype == 'generic':
            self.Fyu_ = Fyu
            self.dFdy_ = dFdy
            self.operator = operator
        elif targettype == 'expectationvalue':
            if isinstance(operator, list):
                self.operator = []
                for i in range(len(operator)):
                    self.operator.append(operator[i].copy())
            else:
                self.operator = [operator.copy()]
        elif targettype == 'evolutionoperator':
            if isinstance(Utarget, list):
                self.Utarget = []
                for i in range(len(Utarget)):
                    self.Utarget.append(Utarget[i].copy())
            else:
                self.Utarget = [Utarget.copy()]
            self.operator = operator
        elif targettype == 'floquet':
            if targeteps is not None:
                self.targeteps = targeteps.copy()
            else:
                self.targeteps = None
            self.T = T
            self.operator = operator
            self.fepsilon_ = fepsilon
            self.dfdepsilon_ = dfdepsilon
        self.dFdu = dFdu
        self.alpha = alpha
        if S == None:
            self.S = lambda x: 1
        else:
            self.S = S


    def Fyu(self, y, u):
        """The functional F of the trayectory y that is to be maximized
        
        It may also be an explicit function of the control parameters u.

        Parameters
        ----------
        y : qutip.result 
            The trajectory of the quantum system
        u : ndarray
            The control parameters that were used to generate the trajectory.

        Returns
        -------
        float:
            The value of the target functional.
        """
        if self.targettype == 'generic':
            if self.dFdy_ == 'ad':
                return float(self.Fyu_(jnp.array(y.full()), u))
            else:
                return self.Fyu_(y, u)

        if self.targettype == 'floquet':
            Fval = 1.0
            if not isinstance(y, list):
                UTset_ = [y]
            else:
                UTset_ = y
            nst = len(UTset_)
            dim = UTset_[0].dims[0][0]
            epsilon = np.zeros([nst, dim])
            for k in range(len(UTset_)):
                epsilon[k, :] = floquet.epsi(UTset_[k].full(), self.T)
            return self.f(epsilon)

        elif self.targettype == 'expectationvalue':
            if isinstance(y, list):
                ntgs = len(y)
                x = 0.0
                for j in range(len(y)):
                    x = x + qt.expect(self.operator[j], y[j])
            else:
                ntgs = 1
                x = qt.expect(self.operator[0], y)
            return x / ntgs

        elif self.targettype == 'evolutionoperator':
            if isinstance(y, list):
                ntgs = len(y)
                x = 0.0
                for j in range(ntgs):
                    dim = y[j].shape[0]
                    x = x + (1/dim**2) * (np.absolute(math_extra.frobenius_product(self.Utarget[j], y[j])))**2
                x = x / ntgs
                return x
            else:
                dim = y.shape[0]
                return (1/dim**2)*(np.absolute(math_extra.frobenius_product(self.Utarget[0], y)))**2

    def dFdy(self, y, u):
        """Derivative of the target functional wrt the quantum trajectory.

        Right now, it in fact assumes that the functional only depends on the final
        state of the quantum trajectory. This functional derivative is used to determine
        the boundary condition used in the definition of the costate.

        Parameters
        ----------
        y: qutip.Qobj or list of qutip.Qobj
           The state of the system at the final time of the propagation
        u: ndarray
           The control parameters

        Returns
        -------
        qutip.Qobj or list of qutip.Qobj
           The derivative of F with respect to the quantum state.
        """
        if self.targettype == 'generic':
            if self.dFdy_ == 'ad':
                shape = np.array(y.full().shape)
                d = shape.prod()
                def Fyureal(yr, u):
                    yreal = yr[0:d]
                    yimag = yr[d:]
                    y = yreal + 1j*yimag
                    y = y.reshape(shape)
                    res = self.Fyu_(y, u)
                    return res
                dFdyureal = grad(Fyureal, argnums = 0)
                # This should put first the real parts, and then the imaginary parts (check!)
                yreal = jnp.hstack([y.full().real.flatten(), y.full().imag.flatten()])
                gf = dFdyureal(yreal, u)
                gfr = gf[:d]
                gfi = gf[d:]
                gfw = (1.0/2.0) * (gfr + 1j * gfi)
                gfw = gfw.reshape(shape)
                gfw = np.array(gfw)
                res = qt.Qobj(gfw)
                return res
            else:
                return self.dFdy_(y, u)

        if self.targettype == 'floquet':
            if not isinstance(y, list):
                UTset_ = [y]
            else:
                UTset_ = y
            nst = len(UTset_)
            dim = UTset_[0].dims[0][0]
            epsilon = np.zeros([nst, dim])
            dfdy = []
            for k in range(len(UTset_)):
                dfdy.append(np.zeros([dim, dim], dtype = complex))
                UT = UTset_[k].full()
                epsilon[k, :] = floquet.epsi(UT, self.T)
                depsi = floquet.depsi(UT, self.T)
                for m in range(dim):
                    dfdy[k][:, :] += self.dfdepsilon(epsilon)[k, m] * depsi[m, :, :]
                dfdy[k] = qt.Qobj(dfdy[k])
            return dfdy

        elif self.targettype == 'expectationvalue':
            if isinstance(y, list):
                dF = []
                ntgs = len(y)
                for j in range(len(y)):
                    if y[j].isket:
                        dF.append(self.operator[j] * y[j] / ntgs)
                    else:
                        dF.append(0.5 * self.operator[j] / ntgs)
                return dF
            else:
                if y.isket:
                    return self.operator[0]*y
                else:
                    return 0.5 * self.operator[0]

        elif self.targettype == 'evolutionoperator':
            if isinstance(y, list):
                dF = []
                ntgs = len(y)
                for j in range(ntgs):
                    dim = y[j].shape[0]
                    dF.append( (1/dim**2) * math_extra.frobenius_product(self.Utarget[j], y[j])*self.Utarget[j] / ntgs)
                return dF
            else:
                dim = y.shape[0]
                return (1/dim**2) * math_extra.frobenius_product(self.Utarget[0], y)*self.Utarget[0]


    def f(self, eps):
        if self.fepsilon_ is not None:
            return self.fepsilon_(eps)
        cte = 1.0
        fval = 0.0
        nkpoints = eps.shape[0]
        targete = self.targeteps
        dim = eps.shape[1]
        fval = 0.0
        for k in range(nkpoints):
            for alpha in range(dim):
                fval = fval - cte * (eps[k, alpha] - targete[k, alpha])**2
        return fval


    def dfdepsilon(self, eps):
        if self.dfdepsilon_ is not None:
            return self.dfdepsilon_(eps)
        cte = 1.0
        nkpoints = eps.shape[0]
        targete = self.targeteps
        dim = eps.shape[1]
        dfval = np.zeros((nkpoints, dim))
        for k in range(nkpoints):
            for alpha in range(dim):
                dfval[k, alpha] = - 2.0 * cte * (eps[k, alpha]-targete[k, alpha])
        return dfval



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg as la
import sys
import nlopt
from qutip import *
from time import time as clocktime


# The diff_ridders routine is taken from the derivcheck distribution.

# Derivcheck is robust and very sensitive tester for analytic derivatives.
# Copyright (C) 2017 Toon Verstraelen <Toon.Verstraelen@UGent.be>.
#
# This file is part of Derivcheck.
#
# Derivcheck is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Derivcheck is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --

def diff_ridders(function, origin, stepsize, con=1.4, safe=2.0, maxiter=15):
    """Estimate first-order derivative with Ridders' finite difference method.

    This implementation is based on the one from the book Numerical Recipes. The code
    is pythonized and no longer using fixed-size arrays. Also, the output of the function
    can be an array.

    Parameters
    ----------
    function : function
        The function to be differentiated.
    origin : float
        The point at which must be differentiated.
    stepsize : float
        The initial step size.
    con : float
        The rate at which the step size is decreased (contracted). Must be larger than
        one.
    safe : float
        The safety check used to terminate the algorithm. If Errors between successive
        orders become larger than ``safe`` times the error on the best estimate, the
        algorithm stop. This happens due to round-off errors.
    maxiter : int
        The maximum number of iterations, equals the maximum number of function calls and
        also the highest polynomial order in the Neville method.

    Returns
    -------
    estimate : float
        The best estimate of the first-order derivative.
    error : float
        The (optimistic) estimate of the error on the derivative.

    """
    if stepsize == 0.0:
        raise ValueError('stepsize must be nonzero.')
    if con <= 1.0:
        raise ValueError('con must be larger than one.')
    if safe <= 1.0:
        raise ValueError('safe must be larger than one.')

    con2 = con*con
    table = [[(
        np.asarray(function(origin + stepsize))
        - np.asarray(function(origin - stepsize))
    )/(2.0*stepsize)]]
    estimate = None
    error = None

    # Loop based on Neville's method.
    # Successive rows in the table will go to smaller stepsizes.
    # Successive columns in the table go to higher orders of extrapolation.
    for i in range(1, maxiter):
        # Reduce step size.
        stepsize /= con
        # First-order approximation at current step size.
        table.append([(
            np.asarray(function(origin + stepsize))
            - np.asarray(function(origin - stepsize))
        )/(2.0*stepsize)])
        # Compute higher-orders
        fac = con2
        for j in range(1, i+1):
            # Compute extrapolations of various orders, requiring no new
            # function evaluations. This is a recursion relation based on
            # Neville's method.
            table[i].append((table[i][j-1]*fac - table[i-1][j-1])/(fac-1.0))
            fac = con2*fac

            # The error strategy is to compare each new extrapolation to one
            # order lower, both at the present stepsize and the previous one:
            current_error = max(abs(table[i][j] - table[i][j-1]).max(),
                                abs(table[i][j] - table[i-1][j-1]).max())

            # If error has decreased, save the improved estimate.
            if error is None or current_error <= error:
                error = current_error
                estimate = table[i][j]

        # If the highest-order estimate is growing larger than the error on the best
        # estimate, the algorithm becomes numerically instable. Time to quit.
        if abs(table[i][i] - table[i-1][i-1]).max() >= safe * error:
            break
        i += 1
    return estimate, error


def timegrid(H0, T, delta):
    """ Generates a time grid in the interval [0, T]"""
    dt0 = delta/H0.norm(norm = 'fro')
    ntsteps_ = T / dt0
    ntsteps = int(ntsteps_)
    time = np.linspace(0, T, ntsteps + (ntsteps+1)%2 )
    return time

def rk4(xi, eom, t, dt):
    k1 = dt * eom(t, xi)
    k2 = dt * eom(t + 0.5*dt, xi + 0.5*k1)
    k3 = dt * eom(t + 0.5*dt, xi + 0.5*k2)
    k4 = dt * eom(t + dt, xi + k3)
    return xi + (1.0/6.0)*(k1 + 2.0*k2 + 2.0*k3 + k4)


def infidelity(U, dmi, dmo):
    """ Returns a measure of the infidelity of a quantum process.

    (...)
    """
    n = len(dmi)
    infs = np.zeros(n)
    infidelity = 0.0
    for j in range(n):
        a = U * dmi[j] * U.dag()
        b = dmo[j]
        dist = ((a-b).dag() * (a-b)).tr()
        infs[j] = dist
        infidelity += dist
    return infidelity/n, np.max(infs)


def dmset(dim, n):
    dm = []
    if n == 2:
        rho1diag = np.zeros(dim)
        for j in range(dim):
            rho1diag[j] = 2*(dim-j) / (dim * (dim+1))
        dm.append( Qobj(np.diag(rho1diag)) )
        dm.append( Qobj(np.ones((dim, dim))/dim) )
    elif n == 3:
        rho1diag = np.zeros(dim)
        for j in range(dim):
            rho1diag[j] = 2*(dim-j) / (dim * (dim+1))
        dm.append( Qobj(np.diag(rho1diag)) )
        dm.append( Qobj(np.ones((dim, dim))/dim) )
        dm.append( qeye(dim)/dim )
    elif n == (dim+1):
        for j in range(dim):
            dm.append( fock_dm(dim, j) )
        dm.append( Qobj(np.ones((dim, dim))/dim) )
    elif n == dim*dim:
        for j in range(dim):
            dm.append(basis(dim, j) * basis(dim, j).dag())
        for j in range(dim):
            for k in range(j+1, dim):
                vec1 = basis(dim, j) + basis(dim, k)
                vec2 = basis(dim, j) + 1j * basis(dim, k)
                dm.append( 0.5 * vec1 * vec1.dag() )
                dm.append( 0.5 * vec2 * vec2.dag() )
    return dm

    
def frobenius_product(A, B):
    """return the Frobenius product between the operators A and B"""
    if type(A) is qutip.qobj.Qobj:
        return (A.dag()*B).tr()
    else:
        return (np.matmul( A.transpose().conjugate(), B)).trace()

uvals = []
maxval = -sys.float_info.max
counter = 0
maxu = 0
nprops = 0
convergence = []
rescode =  { nlopt.SUCCESS: "Success", 
             nlopt.STOPVAL_REACHED: "Stop value reached",
             nlopt.FTOL_REACHED: "Function tolerance reached",
             nlopt.XTOL_REACHED: "Value tolerance reached",
             nlopt.MAXEVAL_REACHED: "Maximum evaluations reached",
             nlopt.MAXTIME_REACHED: "Maximum execution time reached" }

def maximize(func, u0, 
             ftol_abs = 1.0e-6, 
             maxeval = 10,
             stopval = None,
             algorithm = nlopt.LD_LBFGS,
             local_algorithm = None,
             upper_bounds = None,
             lower_bounds = None,
             equality_constraints = None,
             verbose = False,
             of = sys.stdout):
    global maxval
    global counter
    global maxu
    global uvals
    global nprops
    global convergence
    maxval = -sys.float_info.max
    counter = 0
    maxu = 0
    uvals = []
    nprops = 0
    convergence = []
    def wrapper_function(u, grad):
        global uvals
        global maxval
        global counter
        global maxu
        global convergence
        global nprops
        t0 = clocktime()
        uvals.append(u.copy())
        val = func(u, grad)
        if grad.size > 0:
            nprops = nprops + 2
        else:
            nprops = nprops + 1
        if val > maxval:
            maxval = val
            u0 = u.copy()
            maxu = counter
        t1 = clocktime()
        if verbose and of is not None:
            of.write("{:d} {:f} ({:f} s)\n".format(counter, val, t1-t0))
            of.flush()
        convergence.append([counter, nprops, val, t1-t0])
        counter = counter + 1
        return val
    dim = u0.size
    if local_algorithm == None:
        opt = nlopt.opt(algorithm, dim)
    else:
        local_opt = nlopt.opt(local_algorithm, dim)
        local_opt.set_maxeval(maxeval)
        local_opt.set_ftol_abs(ftol_abs)
        opt = nlopt.opt(algorithm, dim)
        opt.set_local_optimizer(local_opt)
    opt.set_max_objective(wrapper_function)
    opt.set_ftol_abs(ftol_abs)
    opt.set_maxeval(maxeval)
    if upper_bounds is not None:
        opt.set_upper_bounds(upper_bounds)
    if lower_bounds is not None:
        opt.set_lower_bounds(lower_bounds)
    if equality_constraints is not None:
        for constraint in equality_constraints:
            opt.add_equality_constraint(constraint[0], constraint[1])
    if stopval is not None:
        opt.set_stopval(stopval)
    try:
        x = opt.optimize(u0)
    except RuntimeError:
        if of is not None:
            of.write("Runtime error\n")
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
    except ValueError:
        if of is not None:
            of.write("Invalid arguments\n")
        result = opt.last_optimize_result()
        x = u0.copy()
        optimum = maxval
    except MemoryError:
        if of is not None:
            of.write("Memory error\n")
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
    except nlopt.RoundoffLimited:
        if of is not None:
            of.write("Round-off error\n")
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
    except nlopt.ForcedStop:
        if of is not None:
            of.write("Forced stop\n")
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
    except:
        if of is not None:
            of.write("Unknown error {}".format(sys.exc_info()[0]))
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
    else:
        result = opt.last_optimize_result()
        x = uvals[maxu]
        optimum = maxval
        if verbose and of is not None: of.write ("Successfull termination with result code = {:d}\n".format(result))
        if verbose and of is not None: of.write('("{}")\n'.format(rescode[result]))
    return x, optimum, convergence, result




## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import scipy as sp
import sys
import nlopt
from qutip import *
from time import time as clocktime
from qocttools.solvers import solve, intoper, solvep
import qocttools.math_extra as math_extra
import qocttools.target as target
import qocttools.krotov as krotov
import qocttools.hamiltonians as hamiltonians
import qocttools.pulses as pulses
import qocttools.floquet as floquet

"""qoct is the main module of the qocttools package

It holds the Qoct class, that is used to setup the optimization
problem.

"""

of = sys.stdout
comm = None
rank = 0

class Qoct:
    """The class that is used to setup the optimization.

    It stores the Hamiltonian, the definition of the target,
    the pulses, the initial state, and some parameters that determine
    how the optimization is done.

    Parameters
    ----------
    H : hamiltonian or list of hamiltonian
        A hamiltonian class instance, or list of instances for multitarget optimizations.
    T : float
        total propagation time
    ntsteps : int
        number of time steps
    tg : Target
        The object that defines the target functional.
    f : pulse or list of pulse
        The pulse object, or list of pulse objects (in the case the Hamiltonian(s) have
        more than one perturbations)
    y0 : Qobj or list of Qobj
        The initial state(s) for the optimization. It should be a list for multitarget
        optimizations. It can be a Hilbert space state, a density matrix, or a propagator.
    interaction_picture : bool, default = False
        The calculations can be done in the interaction picture, or in Schrödinger's picture.
    solve_method: str, default = 'cfmagnus4'
        The method to use for the propagations.
    equality_constraints : list, default = None
        The optimization can be done with constraints, that should be specified as
        a list of [func, tol] objects: func is a function that defines the constraint,
        and tol is a float that sets the tolerance accepted for the fulfillment of that
        constraint.
    sparse : bool, default = False
        It forces the use of sparse algebra only (otherwise, some parts of the calculation
        are done with dense matrices).
    output_file : file, default = sys.stdout
        One may ask the internal procedures of the qoct object to output messages to a
        file different from standard output

    Attributes
    ----------

    y0 : list of Qobj
    H : list of hamiltonian

    """
    def __init__(self, H, T, ntsteps, tg, f, y0,
                 interaction_picture = False,
                 solve_method = 'cfmagnus4',
                 equality_constraints = None,
                 sparse = False, 
                 output_file = sys.stdout,
                 new_parametrization = False,
                 mpi_comm = None,
                 floquet_mode = 'qoct'):

        global of
        global comm
        global rank

        # If y0 and H are not lists, lets us make them lists (with equal number of elements).
        if isinstance(y0, list):
            self.y0 = y0
        else:
            self.y0 = [y0]
        if isinstance(H, list):
            if not (len(H) == len(self.y0)):
                raise Exception("The number of Hamiltonians should be equal to the number of initial states")
            else:
                self.H = H
        else:
            # Normally, all Hamiltonians are equal.
            self.H = []
            for j in range(len(self.y0)):
                (self.H).append(H)
        self.T = T
        self.ntsteps = ntsteps
        self.time = np.linspace(0, T, ntsteps)
        if tg.targettype == 'generic':
            self.O = tg.operator
        elif tg.targettype == 'expectationvalue':
            self.O = tg.operator
        elif tg.targettype == 'evolutionoperator':
            self.O = tg.Utarget
        self.tg = tg

        if isinstance(f, list):
            self.f = []
            for j in range(len(f)):
                (self.f).append(f[j])
        else:
            self.f = [f]
        self.solve_method = solve_method
        self.interaction_picture = interaction_picture

        # The equality constraints
        eqc = []
        eq_tol = 1.0e-6
        for j in range(len(self.f)):
            g = pulses.pulse_constraint_functions(self.f[j], pulses.pulse_collection_parameter_range(self.f, j))
            for k in range(len(g)):
                eqc.append([g[k], eq_tol])
        if len(eqc) > 0:
            self.equality_constraints = []
            for cs in eqc:
                self.equality_constraints.append(cs)
            if equality_constraints is not None:
                for cs in equality_constraints:
                    self.equality_constraints.append(cs)
        else:
            self.equality_constraints = equality_constraints

        self.sparse = sparse
        self.of = output_file
        of = output_file
        self.convergence = None
        self.nprops = 0
        self.optimum = None
        self.new_parametrization = new_parametrization
        comm = mpi_comm
        if comm is not None:
            self.nprocs = comm.Get_size()
            rank = comm.Get_rank()
            if rank == 0:
                of.write("Number of processors = {}\n".format(self.nprocs))
            comm.Barrier()
        else:
            self.nprocs = 1
            rank == 0
        if not (tg.targettype == 'floquet'):
             self.floquet_mode = None
             self.nessopt = None
        else:
             self.floquet_mode = floquet_mode
             if floquet_mode == 'ness':
                  self.nessopt = floquet.Nessopt(tg.operator, T, ntsteps, hamiltonians.toliouville(H), f)
             else:
                  self.nessopt = None


    def almost_converged(self, fraction = 0.99):
        """Returns the iteration number at which the convergence was almost achieved

        It analyzes the convergence history of an optimization, and looks for the
        iteration at which the value of the functional was already a fraction (0.99
        by default) of the maximum.

        Parameters
        ----------
        fraction : float, default = 0.99
            The fraction of the maximum value at which the process is considered
            to be almost converged. If, at iteration 76, the value of the
            functional is already fraction * maximum, then this function
            will return 76 (along with the number of propagations necessary
            to reach iteration 76.

        Returns
        -------
        float
            The iteration at which the process was almost converged
        float
            The number of propagations done until that iteration.
        """
        for iter in self.convergence:
            if(iter[2] >= fraction * self.optimum):
                break
        return iter[0], iter[1]



    def gfunc(self, u):
        """Returns the value of the target functional for a set of parameters u.

        It just calls the gfunction function, using all the information contained
        in the class object.

        Parameters
        ----------
        u : ndarray
            A numpy array holding all the control parameters

        Returns
        -------
        float
            The value of the target parameter.
        """
        tg = self.tg
        Ffunction = tg.Fyu
        H = self.H
        f = self.f
        psi0 = self.y0
        time = self.time
        nessopt = self.nessopt

        if self.floquet_mode == 'pt':
            nkpoints = len(H)
            dim = H[0].dim
            eps = np.zeros([nkpoints, dim])
            for k in range(nkpoints):
                eps[k, :] = floquet.epsilon3(H[k], f, u, time[-1])
            return tg.f(eps)
        elif self.floquet_mode == 'ness':
            return nessopt.G(u)

        if len(psi0) > 1:
            multitarget = True
        else:
            multitarget = False

        pulses.pulse_collection_set_parameters(f, u)

        result = solvep(self.solve_method, H, f, psi0, time,
                        returnQoutput = False, interaction_picture = self.interaction_picture,
                        options = Options(normalize_output = False),
                        comm = comm)
        dim = len(psi0)
        result_states = []
        for i in range(dim):
            result_states.append(Qobj(result[i][time.size-1]))

        return Ffunction(result_states, u) if multitarget else Ffunction(result_states[0],u)


    def dgfunc(self, u):
        """Returns the value and gradient of the target functional for a set of parameters u.


        It just calls the dgfunction function, using all the information contained
        in the class object.

        Parameters
        ----------
        u : ndarray
            A numpy array holding all the control parameters

        Returns
        -------
        float
            The value of the target parameter.
        ndarray
            The gradient of the target functional.
        """

        solve_method = self.solve_method
        tg = self.tg
        Ffunction = tg.Fyu
        dFfunction = tg.dFdy
        H = self.H
        f = self.f
        psi0 = self.y0
        time = self.time
        floquet_mode = self.floquet_mode
        nessopt = self.nessopt
        sparse = self.sparse
        interaction_picture = self.interaction_picture
        dFdu = tg.dFdu
        new_parametrization = self.new_parametrization

        if floquet_mode == 'pt':
            nkpoints = len(H) # This should be generalized eventually.
            dim = H[0].dim
            eps = np.zeros([nkpoints, dim])
            for k in range(nkpoints):
                eps[k, :] = floquet.epsilon3(H[k], f, u, time[-1])
            dfval = tg.dfdepsilon(eps)
            deps = np.zeros((nkpoints, dim, u.shape[0]))
            for k in range(nkpoints):
                deps[k, :, :] = floquet.gradepsilon(H[k], f, u, time[-1])
            val = np.zeros((u.shape[0]))
            for m in range(u.shape[0]):
                val[m] = 0.0
                for k in range(nkpoints):
                    for alpha in range(dim):
                        val[m] = val[m] + dfval[k, alpha] * deps[k, alpha, m]
            return tg.f(eps), val

        elif floquet_mode == 'ness':
            return nessopt.gradG(u)

        # # All the Hamiltonians must be either functions or Qobj's.
        # if H[0].function:
        #      return dgfunction_h_is_function(self, u)

        if len(psi0) > 1:
            multitarget = True
        else:
            multitarget = False

        pulses.pulse_collection_set_parameters(f, u)

        # What are we propagating?
        obj = None
        if psi0[0].type == 'oper':
            if H[0].A is not None:
                obj = 'density'
            else:
                obj = 'propagator'
        else:
            obj = 'state'

        if obj is None:
            raise Exception('Do not know what is the object to propagate')
    
        dim = len(psi0)

        # We propagate the state(s) psi0 to know it in every instant of time
        if sparse:
            result = solvep(solve_method, H, f, psi0, time,
                            returnQoutput = True, interaction_picture = interaction_picture,
                            options = Options(normalize_output = False),
                            comm = comm)
        else:
            result = solvep(solve_method, H, f, psi0, time,
                            returnQoutput = False, interaction_picture = interaction_picture,
                            options = Options(normalize_output = False),
                            comm = comm)

        state_T = [] # States of the system at time T
        for i in range(dim):
            if sparse:
                state_T.append(result[i].states[-1])
            else:
                state_T.append(Qobj(result[i][-1]))

        # Costate(s) of the system at time T. These will be our "initial conditions" for the backpropagation
        costate_T = dFfunction(state_T, u) if multitarget else dFfunction(state_T[0], u)
        if not isinstance(costate_T, list):
            costate_T = [costate_T]

        # Backpropagation
        T = time[time.size - 1]
        tau = T - time #qutip allows backward propagation if an inverted time array is given

        Hclist = []
        for k in range(len(H)):
            if H[k].function:
                Hc = hamiltonians.hamiltonian(H[k].H0, [H[k].V[j] for j in range(len(H[k].V))],
                                              A = H[k].A, 
                                              g = H[k].g)
            else:
                Hc = hamiltonians.hamiltonian(H[k].H0.dag(), [H[k].V[j].dag() for j in range(len(H[k].V))],
                                              A = H[k].A, 
                                              g = H[k].g)
            Hclist.append(Hc)

        if sparse:
            result_costate = solvep(solve_method, Hclist, f, costate_T, tau,
                                    returnQoutput = True, interaction_picture = interaction_picture,
                                    options = Options(normalize_output = False),
                                    comm = comm)
        else:
            result_costate = solvep(solve_method, Hclist, f, costate_T, tau,
                                    returnQoutput = False, interaction_picture = interaction_picture,
                                    options = Options(normalize_output = False),
                                    comm = comm)

        # Calculation of the gradient using the propagated state and the backpropagated costate
        if new_parametrization:
            int_factor = np.zeros([time.size, u.size], dtype = float)
            for c in range(dim):
                int_factor += integration_factor_new(self, u, obj, psi0[c], H[c], result[c], result_costate[c])
        else:
            int_factor = integration_factor(self, u, obj, psi0, H, result, result_costate)

        grad = np.zeros(u.size)
        j = 0
        for k in range(len(f)):
            for m in range(f[k].nu):
                if f[0].type == 'realtime':
                    grad[j] += int_factor[m, j]
                else:
                    grad[j] = 2.0 * sp.integrate.simps(int_factor[:, j], time)
                if dFdu is not None:
                    grad[j] = grad[j] + dFdu(u, j)
                j = j + 1
    
        if multitarget:
            aux=[]
            for c in range (dim):
                if sparse:
                    aux.append(result[c].states[time.size-1])
                else:
                    aux.append(Qobj(result[c][time.size-1]))
            return Ffunction(aux, u), grad
        else:
            if sparse:
                return Ffunction(result[0].states[time.size-1], u), grad
            else:
                return Ffunction(Qobj(result[0][time.size-1]), u), grad


    def check_grad(self, u, m = None, delta = 0.01):
        """A check on the accuracy of the gradient calculation

        It computes one component of the gradient of the target functional using both
        the QOCT expression, and a finite-difference algorithm (the Ridders algorithm).

        Parameters
        ----------
        u : ndarray
            A numpy array holding all the control parameters
        m : int, default = None
            Which gradient component to compute. If None, it will use the one with the
            largest absolute value
        delta : float, default = 0.01
            A parameter determining the starting finite difference in the Ridders
            algorithm.

        Returns
        -------
        float
            The gradient component as computed with the QOCT formula.
        float
            The gradient component as computed with the finite-difference formula.
        float
            An estimation of the gradient accuracy computed with the finite difference formula.
        """
        t0 = clocktime()
        val, grad = self.dgfunc(u)
        t1 = clocktime()
        if rank == 0:
            of.write("Computed the gradient with the QOCT formula. Elapsed time = {} s.\n".format(t1-t0))

        if m is None:
            n = np.argmax(np.abs(grad))
            if rank == 0:
                of.write("Will compute the {} component of the gradient (the largest one).\n".format(n))
                of.flush()
        else:
            n = m
            if rank == 0:
                of.write("Will compute the {} component of the gradient.\n".format(m))
                of.flush()

        def G(x):
            unew = np.copy(u)
            unew[n] = x
            t0 = clocktime()
            gval = self.gfunc(unew)
            t1 = clocktime()
            if rank == 0:
                of.write("G(u) = {} . Elapsed time = {} s.\n".format(gval, t1-t0))
                of.flush()
            return gval
        derivate = math_extra.diff_ridders(G, u[n], delta)
        pulses.pulse_collection_set_parameters(self.f, u)
        return grad[n], derivate[0], derivate[1]


    def maximize(self,
                 maxeval = 100,
                 stopval = None,
                 verbose = False,
                 algorithm = nlopt.LD_LBFGS,
                 local_algorithm = None,
                 upper_bounds = None,
                 lower_bounds = None,
                 tolerance = 1E-06):
        """Performs the QOCT maximization

        This is the main procedure of the class, as it is the one that launches the
        optimization process. There are a number of parameters controlling the way
        in which this optimization is done, that can be controlled here.

        The starting guess parameters for the optimizatio are read from the values
        of the pulses, that were associated to the class object when it was created.

        Parameters
        ----------
        maxeval: int, default = 100
            The maximum number of function evaluations that the optimization algorithm
            is allowed to do. Note that, depending on the algorithm that is used, this 
            may not be strictly enforced.
        stopval: float, default = None
            The optimization will stop if the value of the target function reaches a
            value above stopval.
        verbose: bool, default = False
            The code will output more or less data, depending on the value of this parameter
        algorithm: int, default = nlopt.LD_LBFGS
            The function maximization algorith to be used. qocttools relies on the nlopt
            optimization library:

            https://nlopt.readthedocs.io/en/latest/

            and therefore one can use any of the algorithms that are defined there. See

            https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/

            for a list. If you include the nlopt module in your driver Python code, you
            can use the parameters that are defined there. Although note that in the case
            of Python, the nomenclature would be ``nlopt.LD_MMA``, ``nlopt.LN_COBYLA``, etc., 
            and not ``NLOPT_LD_MMA``, ``NLOPT_LN_COBYLA``, etc. as it is done when using C.

            In addition, if you set ``algorithm = -1``, the code will not use any of the
            nlopt algorithms, but a fairly basic implementation of Krotov's algorithm.
        local_algorithm: int, default = None
            Some nlopt algorithms use a *global* algorithm and a *local* algorithm. If this
            is the case, you need to specify here the local one (whereas the argument of "algorithm"
            would be the global one).
        upper_bounds: ndarray, default = None
            You may specify upper bound for the parameters -- although this is only possible
            with some of the nlopt algorithms; the code will stop if bounds are requested and
            the algorithm is not capable. See the documetation of nlopt for guidance.
        lower_bounds: ndarray, default = None
            The lower bounds for the parameters.
        tolerance: float, default = 1E-06
            The algorithm will consider that a *successful* optimization has been achieved,
            if a certain tolerance is obtained, characterized by this parameter. This is a tolerance
            imposed on the function vale: when the algorithm is not capable of modifying the function
            value beyond this tolerance, the code will stop and, in principle, a local maximum has
            been found. Note, however, that this local maximum may not be a good one. If that is the
            case, one should change the initial guess, the algorithm, or any of the other choices
            that is done.

        Returns
        -------
        ndarray:
            A numpy array with the optimized control parameters.
        float:
            The value of the local maximum that has been achieved
        int:
            An error code; if it is not zero, something went wrong.

        """
        if algorithm == -1:
            x, optimum, self.convergence, result = krotov.krotov(self.solve_method, self.H[0],
                                  self.f, self.y0[0], self.time, self.O[0], self.tg.S,
                                  alpha = self.tg.alpha,
                                  tolerance = tolerance,
                                  verbose = verbose,
                                  maxeval = maxeval,
                                  interaction_picture = self.interaction_picture)
        else:
            def G(u, grad):
                if grad.size > 0:
                    Gval, grad[:] = self.dgfunc(u)
                else:
                    Gval = self.gfunc(u)
                return Gval
            u = pulses.pulse_collection_get_parameters(self.f)
            x, optimum, self.convergence, result = math_extra.maximize(G, u,
                                       maxeval = maxeval,
                                       stopval = stopval,
                                       verbose = verbose,
                                       algorithm = algorithm,
                                       local_algorithm = local_algorithm,
                                       ftol_abs = tolerance,
                                       upper_bounds = upper_bounds,
                                       lower_bounds = lower_bounds,
                                       equality_constraints = self.equality_constraints,
                                       of = self.of)

        if len(self.convergence) > 0:
            self.nprops = self.convergence[-1][1]
        else:
            self.nprops = 0
        self.optimum = optimum
        return x, optimum, result


def integration_factor_new(opt, u, obj, psi0, H, result, result_costate):
    solve_method = opt.solve_method
    f = opt.f
    time = opt.time
    interaction_picture = opt.interaction_picture
    sparse = opt.sparse

    int_factor = np.zeros([time.size, u.size], dtype = float)

    H0 = H.H0
    V = H.V

    n = len(f)
    npert = len(V)
    nu = u.shape[0]

    if not sparse:
        v = []
        for k in range(len(H.V)):
            v.append(V[k].full())
        diagH0 = np.diag(H0.full())

    z = np.zeros(len(V))
    for i in range(time.size):
        #j = 0
        for k in range(len(V)):
            # We calculate the matrix form of the perturbation V

            if sparse:
                pass # No interaction picture allowed here.
            else:
                if interaction_picture:
                    vp = intoper(v[k], diagH0, time[i])
                else:
                    vp = V[k].full()

            # We compute z, which is the matrix element of the costate and the state with V
            if psi0.type == 'ket' or psi0.type == 'bra':
                if sparse:
                    #z = np.imag( result_costate.states[time.size-1-i].dag() * V[k] * result.states[i]   )
                    z[k] = np.imag( result_costate.states[time.size-1-i].overlap( V[k]*result.states[i] ))
                else:
                    z[k] = np.imag( np.vdot(result_costate[time.size-1-i], np.matmul(vp, result[i])) )
            elif psi0.type == 'oper':
                if obj == 'density':
                    p = np.matmul(vp, result[i]) - np.matmul(result[i], vp)
                else:
                    p = np.matmul(vp, result[i])
                z[k] = np.imag( math_extra.frobenius_product(result_costate[time.size - 1 - i], p ))

        ft = np.array( [f[j].fu(time[i]) for j in range(n)] )
        gradf = []
        for j in range(n):
            gradf.append( f[j].gradf(time[i]) )
        for m in range(nu):
            lm = pulses.pulse_collection_l(f, m)
            jm = pulses.pulse_collection_j(f, m)
            int_factor[i, m] = 0.0
            if f[lm].type == 'realtime':
                if m == jm:
                    int_factor[i, m] += 2.0 * z[lm] * time[1]
                    if (i == 0) or (i == time.shape[0]-1):
                        int_factor[i, m] /= 2.0
            else:
                for j in range(npert):
                    int_factor[i, m] = int_factor[i, m] + gradf[lm][jm] * H.gradg(ft, j, lm) * z[j]

    return int_factor


def integration_factor(opt, u, obj, psi0_, H_, result_, result_costate_):
    solve_method = opt.solve_method
    f = opt.f
    time = opt.time
    interaction_picture = opt.interaction_picture
    sparse = opt.sparse

    int_factor = np.zeros([time.size, u.size], dtype = float)

    ns = len(psi0_)

    for n in range(ns):
        psi0 = psi0_[n]
        H = H_[n]
        result = result_[n]
        result_costate = result_costate_[n]

        H0 = H.H0
        V = H.V

        if H.function:
            args = { "f": [f[l].fu for l in range(len(f))] }
        else:
            if not sparse:
                v = []
                for k in range(len(H.V)):
                    v.append(V[k].full())
                diagH0 = np.diag(H0.full())

        for i in range(time.size):
            j = 0
            for k in range(len(V)):
                # We calculate the matrix form of the perturbation V

                if not H.function:
                    if sparse:
                        pass # No interaction picture allowed here.
                    else:
                        if interaction_picture:
                            vp = intoper(v[k], diagH0, time[i])
                        else:
                            vp = V[k].full()

                # We compute z, which is the matrix element of the costate and the state with V
                if psi0.type == 'ket' or psi0.type == 'bra':
                    if H.function:
                        vp = V[k](time[time.size-1-i], args).full()
                        z = np.imag( np.vdot(result_costate[time.size-1-i], np.matmul(vp, result[i])) )
                    else:
                        if sparse:
                            z = np.imag( result_costate.states[time.size-1-i].overlap( V[k]*result.states[i] ))
                        else:
                            z = np.imag( np.vdot(result_costate[time.size-1-i], np.matmul(vp, result[i])) )
                elif psi0.type == 'oper':
                    if H.function:
                        vp = V[k](time[i], args).full()
                        if obj == 'density':
                            p = np.matmul(vp, result[i]) - np.matmul(result[i], vp)
                        else:
                            p = np.matmul(vp, result[i])
                        z = np.imag( math_extra.frobenius_product(result_costate[time.size - 1 - i], p ))
                    else:
                        if obj == 'density':
                            p = np.matmul(vp, result[i]) - np.matmul(result[i], vp)
                        else:
                            p = np.matmul(vp, result[i])
                        z = np.imag( math_extra.frobenius_product(result_costate[time.size - 1 - i], p ))

                gradf = f[k].gradf(time[i])
                for m in range(f[k].nu):
                    if f[k].type == 'realtime': # In this case, the parameters are the amplitude in each time. 
                        if m == i:
                            int_factor[i, j] += 2.0 * z * time[1]
                            if (i == 0) or (i == time.shape[0]-1):
                                int_factor[i, j] /= 2.0
                    else:
                        int_factor[i, j] += gradf[m] * z
                    j = j + 1 # j runs over all the parameters u, and, unless the pulse is 'realtime', j=m

    return int_factor



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module holds a graphene tight-binding model


"""


import numpy as np
import qutip as qt

# There are three constants that must be set for the simple tight-binding model that
# we will be using: t (the hopping parameter), a (the lattice constant) and vf (the Fermi velocity).
# In fact, I think that they are linked by:
#
# vf = 3*a*t/2
#
# I am not sure about the right value for all these constants. I will use the following:
# (in atomic units)

vf = 0.512
a = 2.68
t = 2*vf / (3*a)

a1 = (a/2.0) * np.array([3.0,  np.sqrt(3.0)])
a2 = (a/2.0) * np.array([3.0, -np.sqrt(3.0)])

b1 = (2.0*np.pi/(3.0*a)) * np.array([1, np.sqrt(3)])
b2 = (2.0*np.pi/(3.0*a)) * np.array([1, -np.sqrt(3)])

delta = np.zeros([3, 2])
delta[0, :] = (a/2) * np.array([1.0,  np.sqrt(3.0)])
delta[1, :] = (a/2) * np.array([1.0, -np.sqrt(3.0)])
delta[2, :] = - a * np.array([1.0, 0.0])

Kpoint = (2*np.pi / (3*np.sqrt(3)*a)) * np.array([np.sqrt(3.0), 1.0])
Kpointp = (2*np.pi / (3*np.sqrt(3)*a)) * np.array([np.sqrt(3.0), -1.0])
Mpoint = (2*np.pi / (3*np.sqrt(3)*a)) * np.array([np.sqrt(3.0), 0.0])
Gammapoint = np.array([0.0, 0.0])

def deltak(kvec):
    y = 0j
    for j in range(3):
        y = y + np.exp(1j * np.dot(delta[j], kvec))
    return y

def graddeltak(kvec):
    y = np.zeros(2, dtype = complex)
    for j in range(3):
        y = y + 1j * delta[j, :] * np.exp(1j * np.dot(delta[j], kvec))
    return y

def H0k(kvec):
    dk = deltak(kvec)
    h0kmatrix = np.zeros([2, 2], dtype = complex)
    h0kmatrix[0, 1] = dk
    h0kmatrix[1, 0] = np.conj(dk)
    h0k = qt.Qobj(h0kmatrix)
    return -t * h0k

def H0gappedk(kvec):
    dk = deltak(kvec)
    h0kmatrix = np.zeros([2, 2], dtype = complex)
    h0kmatrix[1, 1] = 0.01/2
    h0kmatrix[0, 0] = -0.01/2
    h0kmatrix[0, 1] = dk
    h0kmatrix[1, 0] = np.conj(dk)
    h0k = qt.Qobj(h0kmatrix)
    return -t * h0k

def Hxk(kvec):
    ggk = graddeltak(kvec)
    mat = np.zeros([2, 2], dtype = complex)
    mat[0, 1] = t * ggk[0]
    mat[1, 0] = np.conj(mat[0, 1])
    return qt.Qobj(mat)

def Hyk(kvec):
    ggk = graddeltak(kvec)
    mat = np.zeros([2, 2], dtype = complex)
    mat[0, 1] = t * ggk[1]
    mat[1, 0] = np.conj(mat[0, 1])
    return qt.Qobj(mat)



